#
# Copyright 2018-2023 Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Logging in to a Renku deployment.

Description
~~~~~~~~~~~

Authenticate with a remote Renku deployment. This command will bring up
a browser window where you can log in using your credentials. Renku CLI
receives and stores a secure token that will be used for future authentications.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.login:login
   :prog: renku login
   :nested: full

Examples
~~~~~~~~

.. code-block:: console

    $ renku login <endpoint>

Parameter ``endpoint`` is the URL of the Renku deployment that you want to
authenticate with (e.g. ``renkulab.io``). You can either pass this parameter on
the command-line or set it once in project's configuration:

.. code-block:: console

    $ renku config set endpoint <endpoint>

.. note::

    The secure token is stored in plain-text in Renku's global configuration
    file on your home directory (``~/.renku/renku.ini``). Renku changes access
    rights of this file to be readable only by you. This token exists only on
    your system and won't be pushed to a remote server.

By default, this command also logs you into gitlab server for project's
repository. You can use this method instead of creating an SSH key. This will
change the repository's remote URL to an endpoint in the deployment that adds
authentication to gitlab requests. Renku warns you each time that the remote
URL will change. To disable this warning, either pass ``--yes`` to the command
or set ``show_login_warning`` to ``false`` for the project or globally:

.. code-block:: console

    $ renku config set [--global] show_login_warning false

You can avoid the remote from being changed by passing ``--no-git`` option to
the login command.

.. note::

    Project's remote URL will be changed when logging in to Renku. Changes
    are undone when logging out from renku in the CLI. Original remote URL will
    be stored in a remote with name ``renku-backup-<remote-name>``.

Logging out from Renku removes the secure token from your system:

.. code-block:: console

    $ renku logout <endpoint>

If you don't specify an endpoint when logging out, credentials for all
endpoints are removed.

.. cheatsheet::
   :group: Misc
   :command: $ renku login --endpoint <URL>
   :description: Login to a Renku deployment for accessing private projects and dataset.
   :target: rp

.. cheatsheet::
   :group: Misc
   :command: $ renku logout --endpoint <URL>
   :description: Logout from a Renku deployment and clear locally-stored credentials.
   :target: rp

"""

import click

import renku.ui.cli.utils.color as color
from renku.ui.cli.utils.callback import ClickCallback


@click.command()
@click.argument("endpoint", required=False, default=None)
@click.option("--no-git", is_flag=True, default=False, help="Don't log in to gitlab.")
@click.option("--yes", is_flag=True, default=False, help="Do not warn users about remote URL change.")
def login(endpoint, no_git, yes):
    """Log in to the platform."""
    from renku.command.login import login_command

    communicator = ClickCallback()
    login_command().with_communicator(communicator).build().execute(endpoint=endpoint, git_login=not no_git, yes=yes)
    click.secho("Successfully logged in.", fg=color.GREEN)


@click.command()
@click.argument("endpoint", required=False, default=None)
def logout(endpoint):
    """Logout from the platform and delete credentials."""
    from renku.command.login import logout_command

    communicator = ClickCallback()
    logout_command().with_communicator(communicator).build().execute(endpoint=endpoint)
    click.secho("Successfully logged out.", fg=color.GREEN)


@click.command(hidden=True)
@click.option("--hostname", default=None, hidden=True, help="Remote hostname.")
@click.argument("command")
def credentials(command, hostname):
    """A git credential helper for returning renku user/token."""
    from renku.command.login import credentials_command

    communicator = ClickCallback()
    credentials_command().with_communicator(communicator).build().execute(command=command, hostname=hostname)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service api versions."""
from typing import Any, Callable, List, Optional

from flask import Blueprint


class ApiVersion:
    """Represents a Blueprint API version."""

    def __init__(self, name: str, is_base_version: bool = False):
        self.name = name
        self.is_base_version = is_base_version


class VersionedBlueprint(Blueprint):
    """A Blueprint that supports versioning."""

    def add_url_rule(
        self,
        rule: str,
        endpoint: Optional[str] = None,
        view_func: Optional[Callable] = None,
        provide_automatic_options: Optional[bool] = None,
        versions: Optional[List[ApiVersion]] = None,
        **options: Any,
    ) -> None:
        """Overwrite Blueprint add_url_rule to support versioning."""
        if versions is None:
            return

        for version in versions:
            if version.is_base_version:
                super().add_url_rule(
                    rule, endpoint, view_func, provide_automatic_options=provide_automatic_options, **options
                )

            version_rule = f"/{version.name}{rule}"
            super().add_url_rule(
                version_rule, endpoint, view_func, provide_automatic_options=provide_automatic_options, **options
            )


V1_0 = ApiVersion("1.0")
V1_1 = ApiVersion("1.1")
V1_2 = ApiVersion("1.2")
V1_3 = ApiVersion("1.3")
V1_4 = ApiVersion("1.4")
V1_5 = ApiVersion("1.5")
V2_0 = ApiVersion("2.0", is_base_version=True)

VERSIONS_FROM_V1_5 = [V1_5, V2_0]
VERSIONS_FROM_V1_4 = [V1_4] + VERSIONS_FROM_V1_5
VERSIONS_FROM_V1_1 = [V1_1, V1_2, V1_3] + VERSIONS_FROM_V1_4
ALL_VERSIONS = [V1_0] + VERSIONS_FROM_V1_1

MINIMUM_VERSION = V1_0
MAXIMUM_VERSION = V2_0



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service workflow serializers."""
from enum import Enum

from marshmallow import Schema, fields
from marshmallow_oneofschema import OneOfSchema

from renku.domain_model.dataset import DatasetCreatorsJson
from renku.ui.cli.utils.plugins import get_supported_formats
from renku.ui.service.serializers.common import LocalRepositorySchema, RemoteRepositorySchema
from renku.ui.service.serializers.rpc import JsonRPCResponse


class WorkflowPlansListRequest(LocalRepositorySchema, RemoteRepositorySchema):
    """Request schema for plan list view."""


class AbstractPlanResponse(Schema):
    """Base schema for abstract plan responses."""

    id = fields.String(required=True)
    name = fields.String(required=True)
    description = fields.String()
    type = fields.String()
    created = fields.DateTime()
    creators = fields.List(fields.Nested(DatasetCreatorsJson))
    keywords = fields.List(fields.String())
    touches_existing_files = fields.Boolean()
    duration = fields.Integer(dump_default=None)


class WorflowPlanEntryResponse(AbstractPlanResponse):
    """Serialize a plan to a response object."""

    last_executed = fields.DateTime()
    number_of_executions = fields.Integer()
    children = fields.List(fields.String)


class WorkflowPlansListResponse(Schema):
    """Response schema for plan list view."""

    plans = fields.List(fields.Nested(WorflowPlanEntryResponse), required=True)


class WorkflowPlansListResponseRPC(JsonRPCResponse):
    """RPC response schema for plan list view."""

    result = fields.Nested(WorkflowPlansListResponse)


class WorkflowPlansShowRequest(LocalRepositorySchema, RemoteRepositorySchema):
    """Request schema for plan show view."""

    plan_id = fields.String(required=True)


class AnnotationSchema(Schema):
    """Custom metadata annotation schema."""

    id = fields.String(required=True)
    source = fields.String()
    body = fields.Raw()


class ParameterBaseSchema(Schema):
    """Base schema for parameters."""

    id = fields.String(required=True)
    plan_id = fields.String(required=True)
    name = fields.String(required=True)
    type = fields.String()
    description = fields.String(allow_none=True)
    default_value = fields.String()
    prefix = fields.String(allow_none=True)
    position = fields.Integer(allow_none=True)


class InputSchema(ParameterBaseSchema):
    """Schema for plan input."""

    mapped_to = fields.String(allow_none=True)
    encoding_format = fields.String(allow_none=True)
    exists = fields.Boolean()


class OutputSchema(ParameterBaseSchema):
    """Schema for plan input."""

    mapped_to = fields.String(allow_none=True)
    encoding_format = fields.String(allow_none=True)
    create_folder = fields.Boolean()
    exists = fields.Boolean()
    last_touched_by_this_plan = fields.Boolean()


class ParameterSchema(ParameterBaseSchema):
    """Schema for a plan parameter."""

    pass


class PlanDetailsResponse(AbstractPlanResponse):
    """Schema for Plan details."""

    last_executed = fields.DateTime()
    number_of_executions = fields.Integer()
    full_command = fields.String()
    command = fields.String()
    inputs = fields.List(fields.Nested(InputSchema))
    outputs = fields.List(fields.Nested(OutputSchema))
    parameters = fields.List(fields.Nested(ParameterSchema))
    success_codes = fields.List(fields.Integer())
    latest = fields.String(dump_default=None)
    annotations = fields.List(fields.Nested(AnnotationSchema))


class PlanReferenceSchema(Schema):
    """Schema for a plan reference."""

    id = fields.String(required=True)
    name = fields.String(required=True)
    description = fields.String(allow_none=True)


class ParameterTargetSchema(Schema):
    """Schema for a mapping target."""

    id = fields.String(required=True)
    plan_id = fields.String(required=True)
    name = fields.String(required=True)
    type = fields.String()


class MappingSchema(Schema):
    """Schema for a plan mapping."""

    id = fields.String(required=True)
    plan_id = fields.String(required=True)
    name = fields.String(required=True)
    type = fields.String()
    description = fields.String(allow_none=True)
    default_value = fields.String()
    targets = fields.List(fields.Nested(ParameterTargetSchema))


class LinkSchema(Schema):
    """Schema for a parameter link."""

    id = fields.String(required=True)
    plan_id = fields.String(required=True)
    type = fields.String()
    source_entry = fields.Nested(ParameterTargetSchema, data_key="source")
    sink_entries = fields.List(fields.Nested(ParameterTargetSchema), data_key="sinks")


class CompositePlanDetailsResponse(AbstractPlanResponse):
    """Schema for Plan details."""

    steps = fields.List(fields.Nested(PlanReferenceSchema), data_key="plans")
    mappings = fields.List(fields.Nested(MappingSchema))
    links = fields.List(fields.Nested(LinkSchema))
    latest = fields.String(dump_default=None)
    annotations = fields.List(fields.Nested(AnnotationSchema))


class PlanSuperSchema(OneOfSchema):
    """Combined schema for Plan and CompositePlan."""

    type_schemas = {"Plan": PlanDetailsResponse, "CompositePlan": CompositePlanDetailsResponse}

    def get_obj_type(self, obj):
        """Get type from object."""
        return str(obj.type)


class WorkflowPlansShowResponseRPC(JsonRPCResponse):
    """Response schema for plan show view."""

    result = fields.Nested(PlanSuperSchema)


WorkflowExportFormatEnum = Enum(  # type: ignore
    "WorkflowExportFormatEnum",
    zip(get_supported_formats(), get_supported_formats()),
)


class WorkflowPlansExportRequest(LocalRepositorySchema, RemoteRepositorySchema):
    """Request schema for exporting a plan."""

    plan_id = fields.String(required=True)
    # NOTE: the enum values (not names) are used to deserialize format, the result from deserialize is Enum
    format = fields.Enum(WorkflowExportFormatEnum, missing=getattr(WorkflowExportFormatEnum, "cwl"), by_value=True)
    values = fields.Dict(keys=fields.String(), missing=None)


class WorkflowPlansExportResponseRPC(JsonRPCResponse):
    """Response schema for exporting a plan."""

    result = fields.String()



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service project edit controller."""
from typing import Dict, cast

from renku.command.project import edit_project_command
from renku.domain_model.constant import NO_VALUE
from renku.ui.service.cache.models.job import Job
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOpSyncMixin
from renku.ui.service.serializers.project import ProjectEditRequest, ProjectEditResponseRPC
from renku.ui.service.views import result_response


class ProjectEditCtrl(ServiceCtrl, RenkuOpSyncMixin):
    """Controller for project edit endpoint."""

    REQUEST_SERIALIZER = ProjectEditRequest()
    RESPONSE_SERIALIZER = ProjectEditResponseRPC()

    def __init__(self, cache, user_data, request_data, migrate_project=False):
        """Construct a project edit controller."""
        self.ctx = cast(Dict, ProjectEditCtrl.REQUEST_SERIALIZER.load(request_data))

        if self.ctx.get("commit_message") is None:
            self.ctx["commit_message"] = "service: project edit"

        super().__init__(cache, user_data, request_data, migrate_project=migrate_project)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""
        if "description" in self.ctx:
            description = self.ctx.get("description")
        else:
            description = NO_VALUE

        if "creator" in self.ctx:
            creator = self.ctx.get("creator")
        else:
            creator = NO_VALUE

        if "custom_metadata" in self.ctx:
            custom_metadata = self.ctx.get("custom_metadata")
        else:
            custom_metadata = NO_VALUE

        if "custom_metadata_source" in self.ctx:
            custom_metadata_source = self.ctx.get("custom_metadata_source")
        else:
            custom_metadata_source = NO_VALUE

        if custom_metadata_source is NO_VALUE and custom_metadata is not NO_VALUE:
            custom_metadata_source = "renku"

        if "keywords" in self.ctx:
            keywords = self.ctx.get("keywords")
        else:
            keywords = NO_VALUE

        result = (
            edit_project_command()
            .with_commit_message(self.ctx["commit_message"])
            .build()
            .execute(
                description=description,
                creator=creator,
                custom_metadata=custom_metadata,
                custom_metadata_source=custom_metadata_source,
                keywords=keywords,
            )
        )

        edited, warning = result.output
        return edited, warning

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        op_result, remote_branch = self.execute_and_sync()

        if isinstance(op_result, Job):
            return result_response(ProjectEditCtrl.JOB_RESPONSE_SERIALIZER, op_result)

        edited, warning = op_result
        response = {
            "edited": edited,
            "warning": warning,
            "remote_branch": remote_branch,
        }

        return result_response(ProjectEditCtrl.RESPONSE_SERIALIZER, response)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service project config views."""
from flask import request

from renku.ui.service.config import SERVICE_PREFIX
from renku.ui.service.controllers.config_set import SetConfigCtrl
from renku.ui.service.controllers.config_show import ShowConfigCtrl
from renku.ui.service.views.api_versions import ALL_VERSIONS, VersionedBlueprint
from renku.ui.service.views.decorators import accepts_json, optional_identity, requires_cache, requires_identity
from renku.ui.service.views.error_handlers import (
    handle_common_except,
    handle_config_read_errors,
    handle_config_write_errors,
)

CONFIG_BLUEPRINT_TAG = "config"
config_blueprint = VersionedBlueprint("config", __name__, url_prefix=SERVICE_PREFIX)


@config_blueprint.route("/config.show", methods=["GET"], provide_automatic_options=False, versions=ALL_VERSIONS)
@handle_common_except
@handle_config_read_errors
@requires_cache
@optional_identity
def show_config(user_data, cache):
    """
    Retrieve the renku config for a project.

    ---
    get:
      description: Retrieve the renku config for a project.
      parameters:
        - in: query
          schema: ConfigShowRequest
      responses:
        200:
          description: Config of a renku project.
          content:
            application/json:
              schema: ConfigShowResponseRPC
      tags:
        - config
    """
    return ShowConfigCtrl(cache, user_data, dict(request.args)).to_response()


@config_blueprint.route("/config.set", methods=["POST"], provide_automatic_options=False, versions=ALL_VERSIONS)
@handle_common_except
@handle_config_write_errors
@accepts_json
@requires_cache
@requires_identity
def set_config(user_data, cache):
    """
    Set the renku config for a project.

    ---
    post:
      description: Set the renku config for a project.
      requestBody:
        content:
          application/json:
            schema: ConfigSetRequest
      responses:
        200:
          description: User and default configuration options.
          content:
            application/json:
              schema: ConfigSetResponseRPC
      tags:
        - config
    """
    return SetConfigCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore



#
# Copyright 2017-2023- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Plugin hooks for workflow file parsers."""

from pathlib import Path
from typing import TYPE_CHECKING, List, Tuple, Union

import pluggy

from renku.core import errors
from renku.core.interface.workflow_file_parser import IWorkflowFileParser

if TYPE_CHECKING:
    from renku.core.workflow.model.workflow_file import WorkflowFile


hookspec = pluggy.HookspecMarker("renku")


@hookspec
def workflow_file_parser() -> Tuple[IWorkflowFileParser, str]:
    """Plugin Hook to get workflow file parsers.

    Returns:
        Tuple[IWorkflowFileParser,str]: A tuple of the parser itself and its name.
    """
    raise NotImplementedError


@hookspec(firstresult=True)
def parse(path: Union[Path, str]) -> "WorkflowFile":
    """Plugin Hook for parsing workflow files.

    Args:
        path(Union[Path, str]): Path to the workflow file to parse.

    Returns:
        WorkflowFile: The parsed workflow file instance.
    """
    raise NotImplementedError


def get_available_workflow_file_parsers() -> List[str]:
    """Returns the currently available workflow file parsers.

    Returns:
        The list of available parsers.
    """
    from renku.core.plugin.pluginmanager import get_plugin_manager

    pm = get_plugin_manager()
    providers = pm.hook.workflow_file_parser()
    return [p[1] for p in providers]


def read_workflow_file(path: Union[Path, str], parser: str = "renku") -> "WorkflowFile":
    """Read a given workflow file using the selected parser.

    Args:
        path(Union[Path, str]): Path to the workflow file.
        parser(str): The workflow parser engine to be used (Default value = "renku").

    Returns:
        WorkflowFile: The parsed workflow file.
    """
    from renku.core.plugin.pluginmanager import get_plugin_manager

    pm = get_plugin_manager()
    parsers = pm.hook.workflow_file_parser()
    selected_parsers = [p for p in parsers if p[1] == parser]

    if not selected_parsers:
        raise errors.ParameterError(f"The specified workflow parser '{parser}' is not available.")
    elif len(parsers) > 1:
        raise errors.ParameterError(f"Multiple parsers found for '{parser}': {selected_parsers}.")

    parsers.remove(selected_parsers[0])
    parse_function = pm.subset_hook_caller("parse", remove_plugins=[p[0] for p in parsers])

    return parse_function(path=path)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku graph endpoints."""
from flask import request

from renku.ui.service.config import SERVICE_PREFIX
from renku.ui.service.controllers.graph_export import GraphExportCtrl
from renku.ui.service.views.api_versions import ALL_VERSIONS, VersionedBlueprint
from renku.ui.service.views.decorators import accepts_json, optional_identity, requires_cache
from renku.ui.service.views.error_handlers import handle_common_except, handle_graph_errors

GRAPH_BLUEPRINT_TAG = "graph"
graph_blueprint = VersionedBlueprint(GRAPH_BLUEPRINT_TAG, __name__, url_prefix=SERVICE_PREFIX)


@graph_blueprint.route("/graph.export", methods=["GET"], provide_automatic_options=False, versions=ALL_VERSIONS)
@handle_common_except
@handle_graph_errors
@requires_cache
@accepts_json
@optional_identity
def graph_build_view(user_data, cache):
    """
    Graph export view.

    ---
    post:
      description: Build a graph for a given repository and revision.
      requestBody:
        content:
          application/json:
            schema: GraphExportRequest
      responses:
        200:
          description: "Status of the graph building"
          content:
            application/json:
              schema: GraphExportResponseRPC
      tags:
        - graph
    """
    return GraphExportCtrl(cache, user_data, dict(request.json)).to_response()  # type: ignore



#
# Copyright 2017-2023 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project management."""

from renku.command.command_builder.command import Command
from renku.core.constant import DATABASE_METADATA_PATH
from renku.core.project import edit_project, show_project


def edit_project_command():
    """Command for editing project metadata."""
    command = Command().command(edit_project).lock_project().with_database(write=True)
    return command.require_migration().with_commit(commit_only=DATABASE_METADATA_PATH)


def show_project_command():
    """Command for showing project metadata."""
    return Command().command(show_project).lock_project().with_database().require_migration()



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache list uploaded files controller."""
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOperationMixin
from renku.ui.service.serializers.cache import FileListResponseRPC
from renku.ui.service.views import result_response


class ListUploadedFilesCtrl(ServiceCtrl, RenkuOperationMixin):
    """Controller for listing uploaded files endpoint."""

    RESPONSE_SERIALIZER = FileListResponseRPC()

    def __init__(self, cache, user_data):
        """Construct list uploaded files controller."""
        self.ctx = {}
        super().__init__(cache, user_data, {})

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""

        files = [f for f in self.cache.get_files(self.user) if f.exists()]
        return {"files": sorted(files, key=lambda rec: (rec.is_dir, rec.relative_path))}

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        return result_response(ListUploadedFilesCtrl.RESPONSE_SERIALIZER, self.renku_op())



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service migrations check controller."""

import tempfile
from dataclasses import asdict
from pathlib import Path

from renku.command.doctor import doctor_check_command
from renku.command.migrate import MigrationCheckResult, migrations_check
from renku.core.errors import AuthenticationError, MinimumVersionError, ProjectNotFound, RenkuException
from renku.core.util.contexts import renku_project_context
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOperationMixin
from renku.ui.service.interfaces.git_api_provider import IGitAPIProvider
from renku.ui.service.serializers.cache import ProjectMigrationCheckRequest, ProjectMigrationCheckResponseRPC
from renku.ui.service.views import result_response


class MigrationsCheckCtrl(ServiceCtrl, RenkuOperationMixin):
    """Controller for migrations check endpoint."""

    REQUEST_SERIALIZER = ProjectMigrationCheckRequest()
    RESPONSE_SERIALIZER = ProjectMigrationCheckResponseRPC()

    def __init__(self, cache, user_data, request_data, git_api_provider: IGitAPIProvider):
        """Construct migration check controller."""
        self.ctx = MigrationsCheckCtrl.REQUEST_SERIALIZER.load(request_data)
        self.git_api_provider = git_api_provider
        super().__init__(cache, user_data, request_data)

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def _fast_op_without_cache(self):
        """Execute renku_op with only necessary files, without cloning the whole repo."""
        if "git_url" not in self.context:
            raise RenkuException("context does not contain `project_id` or `git_url`")

        with tempfile.TemporaryDirectory() as tempdir:
            tempdir_path = Path(tempdir)
            self.git_api_provider.download_files_from_api(
                files=[
                    "Dockerfile",
                ],
                folders=[".renku"],
                target_folder=tempdir_path,
                remote=self.ctx["git_url"],
                branch=self.request_data.get("branch", None),
                token=self.user_data.get("token", None),
            )
            with renku_project_context(tempdir_path):
                return self.renku_op()

    def renku_op(self):
        """Renku operation for the controller."""
        try:
            migrations_check_result = migrations_check().build().execute().output
            doctor_result = doctor_check_command(with_fix=False).build().execute(fix=False, force=False).output
            migrations_check_result.core_compatibility_status.fixes_available = doctor_result[1]
            migrations_check_result.core_compatibility_status.issues_found = doctor_result[2]
            return migrations_check_result
        except MinimumVersionError as e:
            return MigrationCheckResult.from_minimum_version_error(e)

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        from renku.ui.service.views.error_handlers import pretty_print_error

        if "project_id" in self.context:
            result = self.execute_op()
        else:
            # NOTE: use quick flow but fallback to regular flow in case of unexpected exceptions
            try:
                result = self._fast_op_without_cache()
            except (AuthenticationError, ProjectNotFound):
                raise
            except BaseException:
                result = self.execute_op()

        result_dict = asdict(result)

        # NOTE: Pretty-print errors for the UI
        if isinstance(result.template_status, Exception):
            result_dict["template_status"] = pretty_print_error(result.template_status)

        if isinstance(result.dockerfile_renku_status, Exception):
            result_dict["dockerfile_renku_status"] = pretty_print_error(result.dockerfile_renku_status)

        if isinstance(result.core_compatibility_status, Exception):
            result_dict["core_compatibility_status"] = pretty_print_error(result.core_compatibility_status)

        return result_response(self.RESPONSE_SERIALIZER, result_dict)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku modules."""


import importlib
import importlib.util
import sys
import warnings

from renku.version import __template_version__, __version__


class LoaderWrapper(importlib.abc.Loader):
    """Wrap an importlib loader and add the loaded module to sys.modules with an additional name."""

    def __init__(self, wrapped_loader, additional_namespace, original_namespace):
        self.wrapped_loader = wrapped_loader
        self.original_namespace = original_namespace
        self.additional_namespace = additional_namespace

    def get_code(self, fullname):
        """Get executable code.

        Needs to be patched with target name to bypass check in python for names having to match.
        """
        target_name = self.original_namespace + fullname[len(self.additional_namespace) :]
        return self.wrapped_loader.get_code(target_name)

    def create_module(self, spec):
        """Create a module from spec."""
        if spec.name in sys.modules:
            # NOTE: Original name was already loaded, no need for going through all of importlib
            return sys.modules[spec.name]
        return self.wrapped_loader.create_module(spec)

    def exec_module(self, module):
        """Load the module using the wrapped loader."""
        self.wrapped_loader.exec_module(module)
        sys.modules[self.additional_namespace] = module

    def __getattr__(self, name):
        """Forward all calls to wrapped loader except for one implemented here."""
        if name in ["exec_module", "create_module", "get_code"]:
            object.__getattribute__(self, name)

        return getattr(self.wrapped_loader, name)


class DeprecatedImportInterceptor(importlib.abc.MetaPathFinder):
    """Replaces imports of deprecated modules on the fly.

    Replaces old namespace with new namespace, loads the new one, updates sys.modules with the old namespace
    pointing to the new namespace and returns the loaded module.
    """

    def __init__(self, package_redirects):
        self.package_redirects = package_redirects

    def find_spec(self, fullname, path, target=None):
        """Find the spec for a namespace."""
        match = next((n for n in self.package_redirects if fullname.startswith(n)), None)
        if match is not None:
            if match == fullname and self.package_redirects[match][1]:
                warnings.warn(
                    f"The {fullname} module has moved to {self.package_redirects[match][0]} and is deprecated",
                    DeprecationWarning,
                    stacklevel=2,
                )
            try:
                subpath = fullname[len(match) :]
                target_name = self.package_redirects[match][0] + subpath

                sys.meta_path = [x for x in sys.meta_path if x is not self]
                spec = importlib.util.find_spec(target_name)
                if spec is None:
                    return None

                spec.loader = LoaderWrapper(spec.loader, fullname, target_name)
            finally:
                sys.meta_path.insert(0, self)
            return spec

        return None


# NOTE: Patch python import machinery with custom loader
sys.meta_path.insert(
    0,
    DeprecatedImportInterceptor(
        {
            "renku.core.models": ("renku.domain_model", False),
            "renku.core.metadata": ("renku.infrastructure", False),
            "renku.core.commands": ("renku.command", True),
            "renku.core.plugins": ("renku.core.plugin", True),
            "renku.api": ("renku.ui.api", False),
            "renku.cli": ("renku.ui.cli", True),
        }
    ),
)

__all__ = ("__template_version__", "__version__")



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Local provider for local filesystem."""

import os
import urllib
import uuid
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from renku.core import errors
from renku.core.config import get_value
from renku.core.dataset.providers.api import (
    AddProviderInterface,
    ExporterApi,
    ExportProviderInterface,
    ProviderApi,
    ProviderPriority,
)
from renku.core.storage import check_external_storage, track_paths_in_storage
from renku.core.util import communication
from renku.core.util.metadata import is_protected_path
from renku.core.util.os import get_absolute_path, get_safe_relative_path, is_path_empty, is_subpath
from renku.core.util.urls import check_url
from renku.domain_model.project_context import project_context
from renku.infrastructure.immutable import DynamicProxy

if TYPE_CHECKING:
    from renku.core.dataset.providers.models import DatasetAddMetadata, DatasetUpdateMetadata, ProviderParameter
    from renku.domain_model.dataset import Dataset, DatasetTag


class LocalProvider(ProviderApi, AddProviderInterface, ExportProviderInterface):
    """Local filesystem provider."""

    priority = ProviderPriority.LOW
    name = "Local"
    is_remote = False

    def __init__(self, uri: str):
        super().__init__(uri=uri)

        self._path: Optional[str] = None

    @staticmethod
    def supports(uri: str) -> bool:
        """Whether or not this provider supports a given URI."""
        is_remote, _ = check_url(uri)
        return not is_remote

    @staticmethod
    def get_add_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for add."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [
            ProviderParameter(
                "copy",
                flags=["cp", "copy"],
                help="Copy files to the dataset's data directory. Mutually exclusive with --move and --link.",
                is_flag=True,
                default=False,
            ),
            ProviderParameter(
                "move",
                flags=["mv", "move"],
                help="Move files to the dataset's data directory. Mutually exclusive with --copy and --link.",
                is_flag=True,
                default=False,
            ),
            ProviderParameter(
                "link",
                flags=["ln", "link"],
                help="Symlink files to the dataset's data directory. Mutually exclusive with --copy and --move.",
                is_flag=True,
                default=False,
            ),
        ]

    @staticmethod
    def get_export_parameters() -> List["ProviderParameter"]:
        """Returns parameters that can be set for export."""
        from renku.core.dataset.providers.models import ProviderParameter

        return [ProviderParameter("path", flags=["p", "path"], help="Path to copy data to.", type=str)]

    def get_metadata(
        self,
        uri: str,
        destination: Path,
        *,
        move: bool = False,
        copy: bool = False,
        link: bool = False,
        force: bool = False,
        **kwargs,
    ) -> List["DatasetAddMetadata"]:
        """Get metadata of files that will be added to a dataset."""
        from renku.core.dataset.providers.models import DatasetAddAction, DatasetAddMetadata

        repository = project_context.repository

        flags = sum([move, copy, link])
        if flags > 1:
            raise errors.ParameterError("--move, --copy and --link are mutually exclusive.")

        prompt_action = False

        if move:
            default_action = DatasetAddAction.MOVE
        elif link:
            default_action = DatasetAddAction.SYMLINK
        elif copy:
            default_action = DatasetAddAction.COPY
        else:
            prompt_action = True
            action = get_value("renku", "default_dataset_add_action")
            if action:
                prompt_action = False
                if action.lower() == "copy":
                    default_action = DatasetAddAction.COPY
                elif action.lower() == "move":
                    default_action = DatasetAddAction.MOVE
                elif action.lower() == "link":
                    default_action = DatasetAddAction.SYMLINK
                else:
                    raise errors.ParameterError(
                        f"Invalid default action for adding to datasets in Renku config: '{action}'. "
                        "Valid values are 'copy', 'link', and 'move'."
                    )
            else:
                default_action = DatasetAddAction.COPY

        ends_with_slash = False
        u = urllib.parse.urlparse(uri)
        path = u.path

        action = default_action
        source_root = Path(get_absolute_path(path))

        if source_root.is_dir() and uri.endswith("/"):
            ends_with_slash = True

        def check_recursive_addition(src: Path):
            if is_subpath(destination, src):
                raise errors.ParameterError(f"Cannot recursively add path containing dataset's data directory: {path}")

        def get_destination_root():
            destination_exists = destination.exists()
            destination_is_dir = destination.is_dir()

            if is_protected_path(source_root):
                raise errors.ProtectedFiles([source_root])

            check_recursive_addition(source_root)

            if not source_root.exists():
                raise errors.ParameterError(f"Cannot find source file: {path}")
            if source_root.is_dir() and destination_exists and not destination_is_dir:
                raise errors.ParameterError(f"Cannot copy directory '{path}' to non-directory '{destination}'")

            if destination_exists and destination_is_dir:
                if ends_with_slash:
                    return destination

                return destination / source_root.name
            return destination

        def get_file_metadata(src: Path) -> DatasetAddMetadata:
            in_datadir = is_subpath(src, destination)

            relative_path = src.relative_to(source_root)
            dst = destination_root / relative_path

            # NOTE: Add link targets in case they aren't already tracked in the repository
            if action == DatasetAddAction.SYMLINK:
                if check_external_storage():
                    track_paths_in_storage(src)
                repository.add(src)
            source_url = os.path.relpath(src, project_context.path)
            return DatasetAddMetadata(
                entity_path=Path(source_url) if in_datadir else dst.relative_to(project_context.path),
                url=os.path.relpath(src, project_context.path),
                action=DatasetAddAction.NONE if in_datadir else action,
                source=src,
                destination=dst,
            )

        destination_root = get_destination_root()

        if not is_subpath(source_root, project_context.path):
            if link:
                raise errors.ParameterError(f"Cannot use '--link' for files outside of project: '{uri}'")
            if default_action == DatasetAddAction.SYMLINK:
                # NOTE: A default action of 'link' cannot be used for external files
                action = DatasetAddAction.COPY
                prompt_action = True

        results = []
        if source_root.is_dir():
            for file in source_root.rglob("*"):
                if is_protected_path(file):
                    raise errors.ProtectedFiles([file])

                if file.is_dir():
                    check_recursive_addition(file)
                    continue
                results.append(get_file_metadata(file))
        else:
            results = [get_file_metadata(source_root)]

        if not force and prompt_action:
            communication.confirm(
                f"The following files will be copied to {destination.relative_to(project_context.path)}:\n\t"
                "(use '--move' or '--link' to move or symlink them instead, '--copy' to not show this warning).\n\t"
                "(run 'renku config set renku.default_dataset_add_action copy' to make copy the default action).\n\t"
                + "\n\t".join(str(e.source) for e in results)
                + "\nProceed?",
                abort=True,
                warning=True,
            )

        return results

    def update_files(
        self,
        files: List[DynamicProxy],
        dry_run: bool,
        delete: bool,
        context: Dict[str, Any],
        check_data_directory: bool = False,
        **kwargs,
    ) -> List["DatasetUpdateMetadata"]:
        """Update dataset files from the remote provider."""
        from renku.core.dataset.providers.models import DatasetUpdateAction, DatasetUpdateMetadata

        progress_text = "Checking for local updates"
        results: List[DatasetUpdateMetadata] = []

        try:
            communication.start_progress(progress_text, len(files))
            check_paths = []
            records_to_check = []
            for file in files:
                communication.update_progress(progress_text, 1)

                if file.based_on or file.linked:
                    continue

                if not (project_context.path / file.entity.path).exists():
                    results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.DELETE))
                    continue

                check_paths.append(file.entity.path)
                records_to_check.append(file)

            checksums = project_context.repository.get_object_hashes(check_paths)

            for file in records_to_check:
                current_checksum = checksums.get(file.entity.path)
                if not current_checksum:
                    results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.DELETE))
                elif current_checksum != file.entity.checksum:
                    results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.UPDATE))
                elif check_data_directory and not any(file.entity.path == f.entity.path for f in file.dataset.files):
                    datadir = file.dataset.get_datadir()
                    try:
                        get_safe_relative_path(file.entity.path, datadir)
                    except ValueError:
                        continue

                    results.append(DatasetUpdateMetadata(entity=file, action=DatasetUpdateAction.UPDATE))
        finally:
            communication.finalize_progress(progress_text)
        return results

    def get_exporter(
        self, dataset: "Dataset", *, tag: Optional["DatasetTag"], path: Optional[str] = None, **kwargs
    ) -> "LocalExporter":
        """Create export manager for given dataset."""
        self._path = path
        return LocalExporter(dataset=dataset, path=self._path, tag=tag)

    def get_importer(self, uri, **kwargs):
        """Get import manager."""
        raise NotImplementedError


class LocalExporter(ExporterApi):
    """Local filesystem export manager."""

    def __init__(self, dataset: "Dataset", tag: Optional["DatasetTag"], path: Optional[str]):
        super().__init__(dataset)
        self._path: Optional[str] = path
        self._tag: Optional["DatasetTag"] = tag

    @staticmethod
    def requires_access_token() -> bool:
        """Return if export requires an access token."""
        return False

    def set_access_token(self, access_token):
        """Set access token."""
        raise NotImplementedError

    def get_access_token_url(self):
        """Endpoint for creation of access token."""
        return ""

    def export(self, **kwargs) -> str:
        """Execute entire export process."""
        from renku.command.schema.dataset import dump_dataset_as_jsonld
        from renku.core.util.yaml import write_yaml

        if self._path:
            dst_root = project_context.path / self._path
        else:
            dataset_dir = f"{self._dataset.name}-{self._tag.name}" if self._tag else self._dataset.name
            dst_root = project_context.path / project_context.datadir / dataset_dir

        if dst_root.exists() and not dst_root.is_dir():
            raise errors.ParameterError(f"Destination is not a directory: '{dst_root}'")
        elif not is_path_empty(dst_root):
            raise errors.DirectoryNotEmptyError(dst_root)

        dst_root.mkdir(parents=True, exist_ok=True)

        data_dir = self._dataset.get_datadir()

        with communication.progress("Copying dataset files ...", total=len(self._dataset.files)) as progressbar:
            for file in self.dataset.files:
                try:
                    relative_path = str(Path(file.entity.path).relative_to(data_dir))
                except ValueError:
                    relative_path = Path(file.entity.path).name

                dst = dst_root / relative_path
                dst.parent.mkdir(exist_ok=True, parents=True)
                project_context.repository.copy_content_to_file(
                    file.entity.path, checksum=file.entity.checksum, output_path=dst
                )
                progressbar.update()

        metadata_path = dst_root / "METADATA.yml"
        if metadata_path.exists():
            metadata_path = dst_root / f"METADATA-{uuid.uuid4()}.yml"

        metadata = dump_dataset_as_jsonld(self._dataset)
        write_yaml(path=metadata_path, data=metadata)

        communication.echo(f"Dataset metadata was copied to {metadata_path}")
        return str(dst_root)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for renku service dataset controllers."""


def set_url_for_uploaded_images(images, cache, user):
    """Set proper url/path for uploaded or relative dataset images."""
    for img in images:
        if img.get("file_id"):
            file = cache.get_file(user, img.pop("file_id"))
            img["content_url"] = str(file.abs_path)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Remove a file, a directory, or a symlink.

Description
~~~~~~~~~~~

Remove a file from then project, updating:

* its metadata, if the file belongs to a dataset
* the tracking information, if the file is stored in an external storage
  (using Git LFS).

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.remove:remove
   :prog: renku rm
   :nested: full

.. cheatsheet::
   :group: Misc
   :command: $ renku rm <path>...
   :description: Safely delete files from a project.
   :target: rp
"""

import click

from renku.ui.cli.utils.callback import ClickCallback


@click.command(name="rm")
@click.argument("sources", type=click.Path(exists=True), nargs=-1, required=True)
def remove(sources):
    """Remove files and check repository for potential problems."""
    from renku.command.remove import remove_command

    communicator = ClickCallback()
    remove_command().with_communicator(communicator).build().execute(sources=sources, edit_command=click.edit)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service project serializers."""
from marshmallow import fields
from marshmallow.schema import Schema

from renku.domain_model.dataset import DatasetCreatorsJson as DatasetCreators
from renku.ui.service.serializers.common import (
    AsyncSchema,
    LocalRepositorySchema,
    MigrateSchema,
    RemoteRepositoryBaseSchema,
    RemoteRepositorySchema,
    RenkuSyncSchema,
)
from renku.ui.service.serializers.rpc import JsonRPCResponse


class ProjectShowRequest(AsyncSchema, LocalRepositorySchema, RemoteRepositorySchema, MigrateSchema):
    """Project show metadata request."""


class ProjectShowResponse(Schema):
    """Response schema for project show."""

    id = fields.String(metadata={"description": "The ID of this project"})
    name = fields.String(metadata={"description": "The name of the project"})
    description = fields.String(dump_default=None, metadata={"description": "The optional description of the project"})
    created = fields.DateTime(metadata={"description": "The date this project was created at."})
    creator = fields.Nested(DatasetCreators, metadata={"description": "The creator of this project"})
    agent = fields.String(metadata={"description": "The renku version last used on this project"})
    custom_metadata = fields.List(
        fields.Dict(),
        dump_default=None,
        attribute="annotations",
        metadata={"description": "Custom JSON-LD metadata of the project"},
    )
    template_info = fields.String(
        metadata={"description": "The template that was used in the creation of this project"}
    )
    keywords = fields.List(
        fields.String(),
        dump_default=None,
        load_default=None,
        metadata={"description": "They keywords associated with this project"},
    )


class ProjectShowResponseRPC(RenkuSyncSchema):
    """RPC schema for project show."""

    result = fields.Nested(ProjectShowResponse)


class ProjectEditRequest(AsyncSchema, LocalRepositorySchema, RemoteRepositorySchema, MigrateSchema):
    """Project edit metadata request."""

    description = fields.String(metadata={"description": "New description for the project"})
    creator = fields.Nested(DatasetCreators, metadata={"description": "New creator for the project"})
    custom_metadata = fields.List(
        fields.Dict(), metadata={"description": "New custom metadata for the project"}, allow_none=True
    )
    custom_metadata_source = fields.String(
        allow_none=True,
        metadata={"description": "The source for the JSON-LD metadata"},
    )
    keywords = fields.List(fields.String(), allow_none=True, metadata={"description": "New keyword(s) for the project"})


class ProjectEditResponse(RenkuSyncSchema):
    """Project edit metadata response."""

    edited = fields.Dict(required=True, metadata={"description": "Key:value paris of edited metadata"})
    warning = fields.String(metadata={"description": "Warnings raised when editing metadata"})


class ProjectEditResponseRPC(JsonRPCResponse):
    """RPC schema for a project edit."""

    result = fields.Nested(ProjectEditResponse)


class ProjectLockStatusRequest(LocalRepositorySchema, RemoteRepositoryBaseSchema):
    """Project lock status request."""

    timeout = fields.Float(
        dump_default=0.0,
        load_default=0.0,
        metadata={"description": "Maximum amount of time to wait trying to acquire a lock when checking status."},
    )


class ProjectLockStatusResponse(Schema):
    """Project lock status response."""

    locked = fields.Boolean(required=True, metadata={"description": "Whether or not a project is locked for writing"})


class ProjectLockStatusResponseRPC(JsonRPCResponse):
    """RPC schema for project lock status."""

    result = fields.Nested(ProjectLockStatusResponse)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Convenience method to save local changes and push them to a remote server.

Commands and options
~~~~~~~~~~~~~~~~~~~~

.. rst-class:: cli-reference-commands

.. click:: renku.ui.cli.save:save
   :prog: renku save
   :nested: full

Examples
~~~~~~~~

If you have local modification to files, you can save them using

.. code-block:: console

    $ renku save
    Username for 'https://renkulab.io': my.user
    Password for 'https://my.user@renkulab.io':
    Successfully saved:
        file1
        file2
    OK

.. warning:: The username and password for renku save are your gitlab
   user/password, not your renkulab login!

.. cheatsheet::
   :group: Misc
   :command: $ renku save [-m <message>]
   :description: Save (commit) and push all local changes. with optional
                 message.
   :target: rp

You can additionally supply a message that describes the changes that you
made by using the ``-m`` or ``--message`` parameter followed by your
message.

.. code-block:: console

    $ renku save -m "Updated file1 and 2."
    Successfully saved:
        file1
        file2
    OK

If no remote server has been configured, you can specify one by using the
``-d`` or ``--destination`` parameter. Otherwise you will get an error.

.. code-block:: console

    $ renku save
    Error: No remote has been set up for the current branch

    $ renku save -d https://renkulab.io/gitlab/my.user/my-project.git
    Successfully saved:
        file1
        file2
    OK

You can also specify which paths to save:

.. code-block:: console

    $ renku save file1
    Successfully saved:
        file1
    OK

"""

import click

import renku.ui.cli.utils.color as color
from renku.ui.cli.utils.callback import ClickCallback


@click.command(name="save")
@click.option("-m", "--message", default=None, help="The commit message to use")
@click.option(
    "-d",
    "--destination",
    default=None,
    help=(
        "The git remote to push to. Defaults to the remote set in git, " "automatically set in interactive environments"
    ),
)
@click.argument("paths", type=click.Path(exists=False, dir_okay=True), nargs=-1)
def save(message, destination, paths):
    """Save and push local changes."""
    from renku.command.save import save_and_push_command

    communicator = ClickCallback()
    saved_paths, branch = (
        save_and_push_command()
        .with_communicator(communicator)
        .build()
        .execute(message=message, remote=destination, paths=paths)
        .output
    )

    if saved_paths:
        paths_str = "\n\t".join(saved_paths)
        click.echo(f"Successfully saved to remote branch {branch}: \n\t{paths_str}")
    else:
        click.echo(f"Successfully saved to remote branch {branch}.")

    click.secho("OK", fg=color.GREEN)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Datasets JSON-LD schemes."""

from marshmallow import EXCLUDE, pre_dump

from renku.command.schema.agent import PersonSchema
from renku.command.schema.annotation import AnnotationSchema
from renku.command.schema.calamus import DateTimeList, JsonLDSchema, Nested, Uri, fields, oa, prov, renku, schema
from renku.command.schema.entity import CollectionSchema, EntitySchema
from renku.domain_model.dataset import Dataset, DatasetFile, DatasetTag, ImageObject, Language, RemoteEntity, Url


def dump_dataset_as_jsonld(dataset: Dataset) -> dict:
    """Return JSON-LD representation of a dataset.

    Args:
        dataset(Dataset): The dataset to convert.

    Returns:
        dict: JSON-LD data of dataset.

    """
    return DatasetSchema(flattened=True).dump(dataset)


class UrlSchema(JsonLDSchema):
    """Url schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.URL
        model = Url
        unknown = EXCLUDE

    id = fields.Id(load_default=None)
    url = Uri(schema.url, load_default=None)


class DatasetTagSchema(JsonLDSchema):
    """DatasetTag schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.PublicationEvent
        model = DatasetTag
        unknown = EXCLUDE

    dataset_id = Nested(schema.about, UrlSchema, load_default=None)
    date_created = fields.DateTime(schema.startDate, load_default=None, format="iso", extra_formats=("%Y-%m-%d",))
    description = fields.String(schema.description, load_default=None)
    id = fields.Id()
    name = fields.String(schema.name)

    @pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.unfreeze()
            obj.date_created = obj.date_created.replace(microsecond=0)
            obj.freeze()

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs


class LanguageSchema(JsonLDSchema):
    """Language schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.Language
        model = Language
        unknown = EXCLUDE

    alternate_name = fields.String(schema.alternateName)
    id = fields.Id(load_default=None)
    name = fields.String(schema.name)


class ImageObjectSchema(JsonLDSchema):
    """ImageObject schema."""

    class Meta:
        """Meta class."""

        rdf_type = schema.ImageObject
        model = ImageObject
        unknown = EXCLUDE

    content_url = fields.String(schema.contentUrl)
    id = fields.Id(load_default=None)
    position = fields.Integer(schema.position)


class RemoteEntitySchema(JsonLDSchema):
    """RemoteEntity schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Entity]
        model = RemoteEntity
        unknown = EXCLUDE

    checksum = fields.String(renku.checksum)
    id = fields.Id()
    path = fields.String(prov.atLocation)
    url = fields.String(schema.url)


class DatasetFileSchema(JsonLDSchema):
    """DatasetFile schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Entity, schema.DigitalDocument]
        model = DatasetFile
        unknown = EXCLUDE

    based_on = Nested(schema.isBasedOn, RemoteEntitySchema, load_default=None)
    date_added = DateTimeList(schema.dateCreated, format="iso", extra_formats=("%Y-%m-%d",))
    date_removed = fields.DateTime(prov.invalidatedAtTime, load_default=None, format="iso")
    entity = Nested(prov.entity, [EntitySchema, CollectionSchema])
    id = fields.Id()
    is_external = fields.Boolean(renku.external, load_default=False)
    source = fields.String(renku.source, load_default=None)

    @pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.date_added = obj.date_added.replace(microsecond=0)
            if obj.date_removed:
                obj.date_removed = obj.date_removed.replace(microsecond=0)

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs


class DatasetSchema(JsonLDSchema):
    """Dataset schema."""

    class Meta:
        """Meta class."""

        rdf_type = [prov.Entity, schema.Dataset]
        model = Dataset
        unknown = EXCLUDE

    annotations = Nested(oa.hasTarget, AnnotationSchema, reverse=True, many=True)
    creators = Nested(schema.creator, PersonSchema, many=True)
    date_created = fields.DateTime(schema.dateCreated, load_default=None, format="iso", extra_formats=("%Y-%m-%d",))
    date_modified = fields.DateTime(schema.dateModified, load_default=None, format="iso", extra_formats=("%Y-%m-%d",))
    date_removed = fields.DateTime(prov.invalidatedAtTime, load_default=None, format="iso")
    date_published = fields.DateTime(
        schema.datePublished, load_default=None, format="%Y-%m-%d", extra_formats=("iso", "%Y-%m-%dT%H:%M:%S")
    )
    derived_from = Nested(prov.wasDerivedFrom, UrlSchema, load_default=None)
    description = fields.String(schema.description, load_default=None)
    dataset_files = Nested(schema.hasPart, DatasetFileSchema, many=True)
    id = fields.Id(load_default=None)
    identifier = fields.String(schema.identifier)
    images = fields.Nested(schema.image, ImageObjectSchema, load_default=None, many=True)
    in_language = Nested(schema.inLanguage, LanguageSchema, load_default=None)
    keywords = fields.List(schema.keywords, fields.String(), load_default=None)
    license = Uri(schema.license, load_default=None)
    name = fields.String(renku.slug)
    initial_identifier = fields.String(renku.originalIdentifier)
    project_id = fields.IRI(renku.hasDataset, reverse=True)
    same_as = Nested(schema.sameAs, UrlSchema, load_default=None)
    title = fields.String(schema.name)
    version = fields.String(schema.version, load_default=None)

    @pre_dump(pass_many=True)
    def removes_ms(self, objs, many, **kwargs):
        """Remove milliseconds from datetimes.

        Note: since DateField uses `strftime` as format, which only supports timezone info without a colon
        e.g. `+0100` instead of `+01:00`, we have to deal with milliseconds manually instead of using a format string.
        """

        def _replace_times(obj):
            obj.unfreeze()
            if obj.date_created:
                obj.date_created = obj.date_created.replace(microsecond=0)
            if obj.date_removed:
                obj.date_removed = obj.date_removed.replace(microsecond=0)
            if obj.date_published:
                obj.date_published = obj.date_published.replace(microsecond=0)
            obj.date_modified = obj.date_modified.replace(microsecond=0)
            obj.freeze()

        if many:
            for obj in objs:
                _replace_times(obj)
            return objs

        _replace_times(objs)
        return objs



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initial migrations."""

import glob
import os
import traceback
import uuid
from functools import cmp_to_key
from hashlib import sha1
from pathlib import Path
from urllib.parse import urlparse

from cwl_utils.parser import load_document_by_uri
from cwl_utils.parser.cwl_v1_0 import CommandLineTool, Directory, File, InitialWorkDirRequirement
from werkzeug.utils import secure_filename

from renku.core.constant import RENKU_HOME
from renku.core.migration.models.migration import MigrationType, RepositoryCache
from renku.core.migration.models.v3 import Dataset
from renku.core.migration.models.v9 import (
    Collection,
    CommandArgument,
    CommandInput,
    CommandOutput,
    Entity,
    MappedIOStream,
    Person,
    ProcessRun,
    Run,
    SoftwareAgent,
    WorkflowRun,
)
from renku.core.migration.utils import OLD_DATASETS_PATH, OLD_WORKFLOW_PATH
from renku.core.util import communication
from renku.core.util.git import get_in_submodules
from renku.domain_model.project_context import project_context
from renku.infrastructure.repository import Actor, Commit
from renku.version import __version__, version_url

default_missing_software_agent = SoftwareAgent(
    label="renku pre 0.11.0", id="https://github.com/swissdatasciencecenter/renku-python/tree/pre-0.11.0"
)


def migrate(migration_context):
    """Migration function."""
    if MigrationType.WORKFLOWS not in migration_context.options.type:
        return

    migration_context.cwl_cache = {}
    try:
        _migrate_old_workflows(migration_context=migration_context, strict=migration_context.options.strict)
    finally:
        migration_context.cwl_cache = None


def _migrate_old_workflows(migration_context, strict):
    """Migrates old cwl workflows to new jsonld format."""

    def sort_cwl_commits(e1, e2):
        """Sorts cwl commits in order of their creation."""
        commit1 = e1[1]
        commit2 = e2[1]

        return commit1.compare_to(commit2)

    cache = RepositoryCache.from_repository(project_context.repository)
    migration_context.cache = cache

    try:
        wf_path = f"{project_context.metadata_path}/workflow/*.cwl"
        for path in glob.glob(wf_path):
            if path not in cache.cwl_files_commits:
                raise ValueError(f"Couldn't find a previous commit for path `{path}`")

        cwl_paths = list(cache.cwl_files_commits.items())
        cwl_paths = sorted(cwl_paths, key=cmp_to_key(sort_cwl_commits))

        for n, element in enumerate(cwl_paths, start=1):
            communication.echo(f"Processing commit {n}/{len(cwl_paths)}", end="\r")

            cwl_file, commit = element
            repository = project_context.repository

            try:
                if not Path(cwl_file).exists():
                    continue

                path = _migrate_cwl(migration_context=migration_context, path=cwl_file, commit=commit)
                os.remove(cwl_file)

                repository.add(cwl_file, path)

                if repository.is_dirty(untracked_files=False):
                    commit_msg = "renku migrate: committing migrated workflow"
                    committer = Actor(name=f"renku {__version__}", email=version_url)
                    repository.commit(commit_msg + project_context.transaction_id, committer=committer, no_verify=True)
            except Exception:
                if strict:
                    raise
                communication.echo("")
                communication.warn(f"Cannot process commit '{commit.hexsha}' - Exception: {traceback.format_exc()}")
    finally:
        migration_context.cache = None


def _migrate_cwl(migration_context, path, commit):
    """Migrate a cwl file."""
    workflow = parse_cwl_cached(migration_context=migration_context, path=str(path))

    if isinstance(workflow, CommandLineTool):
        _, path = _migrate_single_step(
            migration_context=migration_context, cmd_line_tool=workflow, path=path, commit=commit, persist=True
        )
    else:
        _, path = _migrate_composite_step(
            migration_context=migration_context, workflow=workflow, path=path, commit=commit
        )

    return path


def _migrate_single_step(migration_context, cmd_line_tool, path, commit=None, parent_commit=None, persist=False):
    """Migrate a single step workflow."""
    if not commit:
        commit = migration_context.cache.find_previous_commit(path, revision=parent_commit if parent_commit else "HEAD")

    run = Run(path=path, commit=commit)
    run.command = " ".join(cmd_line_tool.baseCommand)
    run.successcodes = cmd_line_tool.successCodes

    inputs = list(cmd_line_tool.inputs)
    outputs = list(cmd_line_tool.outputs)

    # NOTE: Make run ids deterministic to prevent duplication.
    rel_path = Path(path).relative_to(project_context.path)
    if parent_commit:
        label = f"{rel_path}@{parent_commit.hexsha}"
    else:
        label = f"{rel_path}@{commit.hexsha}"
    identifier = sha1(label.encode("utf-8")).hexdigest()

    base_id = Run.generate_id(identifier=identifier)
    run._id = base_id

    if cmd_line_tool.stdin:
        name = cmd_line_tool.stdin.split(".")[1]

        if name.endswith(")"):
            name = name[:-1]

        matched_input = next(i for i in inputs if i.id.endswith(name))
        inputs.remove(matched_input)

        path = project_context.metadata_path / OLD_WORKFLOW_PATH / Path(matched_input.default.path)
        stdin = path.resolve().relative_to(project_context.path)
        id_ = CommandInput.generate_id(base_id, "stdin")

        run.inputs.append(
            CommandInput(
                id=id_,
                consumes=_entity_from_path(migration_context, stdin, commit),
                mapped_to=MappedIOStream(stream_type="stdin"),
            )
        )

    if cmd_line_tool.stdout:
        run.outputs.append(
            CommandOutput(
                id=CommandOutput.generate_id(base_id, "stdout"),
                produces=_entity_from_path(migration_context, cmd_line_tool.stdout, commit),
                mapped_to=MappedIOStream(stream_type="stdout"),
                create_folder=False,
            )
        )

        matched_output = next(o for o in outputs if o.id.endswith("output_stdout"))

        if matched_output:
            outputs.remove(matched_output)

    if cmd_line_tool.stderr:
        run.outputs.append(
            CommandOutput(
                id=CommandOutput.generate_id(base_id, "stderr"),
                produces=_entity_from_path(migration_context, cmd_line_tool.stderr, commit),
                mapped_to=MappedIOStream(stream_type="stderr"),
                create_folder=False,
            )
        )

        matched_output = next(o for o in outputs if o.id.endswith("output_stderr"))

        if matched_output:
            outputs.remove(matched_output)

    generated_outputs = []
    workdir_requirements = [r for r in cmd_line_tool.requirements if isinstance(r, InitialWorkDirRequirement)]

    for r in workdir_requirements:
        for listing in r.listing:
            if listing.entry == '$({"listing": [], "class": "Directory"})':
                generated_outputs.append(listing.entryname)

    # NOTE: multiple outputs might bind to the same input; we use this copy to find output bindings
    all_inputs = inputs.copy()

    for o in outputs:
        prefix = None
        position = None

        if o.outputBinding.glob.startswith("$(inputs."):
            name = o.outputBinding.glob.split(".")[1]

            if name.endswith(")"):
                name = name[:-1]

            matched_input = next(i for i in all_inputs if i.id.endswith(name))
            try:
                inputs.remove(matched_input)
            except ValueError:
                pass

            if isinstance(matched_input.default, dict):
                path = project_context.metadata_path / OLD_WORKFLOW_PATH / Path(matched_input.default.path)
            else:
                path = Path(matched_input.default)

            path = Path(os.path.realpath(project_context.path / path)).relative_to(project_context.path)

            if matched_input.inputBinding:
                prefix = matched_input.inputBinding.prefix
                position = matched_input.inputBinding.position

                if prefix and matched_input.inputBinding.separate:
                    prefix += " "
        else:
            path = Path(o.outputBinding.glob)

        create_folder = False

        check_path = path
        if not (project_context.path / path).is_dir():
            check_path = path.parent

        if check_path != "." and str(check_path) in generated_outputs:
            create_folder = True

        run.outputs.append(
            CommandOutput(
                id=CommandOutput.generate_id(base_id, position),
                position=position,
                prefix=prefix,
                produces=_entity_from_path(migration_context, path, commit),
                create_folder=create_folder,
            )
        )

    for i in inputs:
        prefix = None
        position = None

        if i.inputBinding:
            prefix = i.inputBinding.prefix
            position = i.inputBinding.position

            if prefix and i.inputBinding.separate:
                prefix += " "

        if isinstance(i.default, (File, Directory)):
            path = project_context.metadata_path / OLD_WORKFLOW_PATH / Path(str(i.default.path))
            path = Path(os.path.realpath(path)).relative_to(project_context.path)

            run.inputs.append(
                CommandInput(
                    id=CommandInput.generate_id(base_id, position),
                    position=position,
                    prefix=prefix,
                    consumes=_entity_from_path(migration_context, path, commit),
                )
            )
        else:
            run.arguments.append(
                CommandArgument(
                    id=CommandArgument.generate_id(base_id, position),
                    position=position,
                    prefix=prefix,
                    value=str(i.default),
                )
            )

    for a in cmd_line_tool.arguments:
        id_ = CommandArgument.generate_id(base_id, a.position)
        run.arguments.append(CommandArgument(id=id_, position=a.position, value=a.valueFrom))

    if not persist:
        return run, None

    step_name = "{}_{}.yaml".format(uuid.uuid4().hex, secure_filename("_".join(cmd_line_tool.baseCommand)))

    absolute_path = project_context.metadata_path / OLD_WORKFLOW_PATH / step_name
    path = absolute_path.relative_to(project_context.path)

    run.path = path
    process_run = ProcessRun.from_run(run, path, commit=commit)
    process_run.invalidated = _invalidations_from_commit(commit)

    # HACK: This fixes broken SoftwareAgent due to re-bases done by users
    if isinstance(process_run.association.agent, Person) or not process_run.association.agent.label.startswith(
        "renku "
    ):
        process_run.association.agent = default_missing_software_agent
    process_run.to_yaml(path=absolute_path)
    return process_run, absolute_path


def _migrate_composite_step(migration_context, workflow, path, commit=None):
    """Migrate a composite workflow."""
    if not commit:
        commit = migration_context.cache.find_previous_commit(path)
    run = Run(path=path, commit=commit)
    rel_path = Path(path).relative_to(project_context.path)
    label = f"{rel_path}@{commit.hexsha}"
    identifier = sha1(label.encode("utf-8")).hexdigest()
    run._id = Run.generate_id(identifier=identifier)

    name = f"{uuid.uuid4().hex}_migrated.yaml"

    wf_path = project_context.metadata_path / OLD_WORKFLOW_PATH
    run.path = (wf_path / name).relative_to(project_context.path)

    for step in workflow.steps:
        if isinstance(step.run, dict):
            continue
        else:
            uri = urlparse(step.run)
            path = uri.path
            sub_run = parse_cwl_cached(migration_context=migration_context, path=path)

        subprocess, _ = _migrate_single_step(
            migration_context=migration_context, cmd_line_tool=sub_run, path=path, parent_commit=commit
        )
        run.add_subprocess(subprocess)

    wf = WorkflowRun.from_run(run, run.path, commit=commit)

    # HACK: This fixes broken SoftwareAgent due to re-bases done by users
    if isinstance(wf.association.agent, Person) or not wf.association.agent.label.startswith("renku "):
        wf.association.agent = default_missing_software_agent
    for p in wf._processes:
        if isinstance(p.association.agent, Person) or not p.association.agent.label.startswith("renku "):
            p.association.agent = default_missing_software_agent
    wf.to_yaml(run.path)

    return wf, run.path


def _entity_from_path(migration_context, path, commit):
    """Gets the entity associated with a path."""
    _, commit, path = get_in_submodules(
        project_context.repository, migration_context.cache.find_previous_commit(path, revision=commit), path
    )

    entity_cls = Entity
    if (project_context.path / path).is_dir():
        entity_cls = Collection

    if str(path).startswith(os.path.join(RENKU_HOME, OLD_DATASETS_PATH)):
        return Dataset.from_yaml(path=project_context.path / path, commit=commit)
    else:
        return entity_cls(commit=commit, path=str(path))


def _invalidations_from_commit(commit: Commit):
    """Gets invalidated files from a commit."""
    results = []
    collections = dict()
    for file in commit.get_changes():
        # NOTE: only process deleted files
        if not file.deleted:
            continue
        path = Path(file.a_path)
        entity = _get_activity_entity(commit, path, collections, deleted=True)

        results.append(entity)

    return results


def _get_activity_entity(commit, path, collections, deleted=False):
    """Gets the entity associated with this Activity and path."""
    _, commit, path = get_in_submodules(repository=project_context.repository, commit=commit, path=path)
    output_path = project_context.path / path
    parents = list(output_path.relative_to(project_context.path).parents)

    collection = None
    members = []
    for parent in reversed(parents[:-1]):
        if str(parent) in collections:
            collection = collections[str(parent)]
        else:
            collection = Collection(commit=commit, path=str(parent), members=[], parent=collection)
            members.append(collection)
            collections[str(parent)] = collection

        members = collection.members

    entity_cls = Entity
    if (project_context.path / path).is_dir():
        entity_cls = Collection

    if str(path).startswith(os.path.join(RENKU_HOME, OLD_DATASETS_PATH)) and not deleted:
        entity = Dataset.from_yaml(path=project_context.path / path, commit=commit)
    else:
        entity = entity_cls(commit=commit, path=str(path), parent=collection)

    if collection:
        collection.members.append(entity)

    return entity


def parse_cwl_cached(migration_context, path: str):
    """Parse cwl and remember the result for future execution."""
    if path in migration_context.cwl_cache:
        return migration_context.cwl_cache[path]

    cwl = load_document_by_uri(path)

    migration_context.cwl_cache[path] = cwl

    return cwl



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache project related models."""
import os
import shutil
from datetime import datetime
from typing import Optional

import portalocker
from walrus import BooleanField, DateTimeField, IntegerField, Model, TextField

from renku.ui.service.cache.base import BaseCache
from renku.ui.service.config import CACHE_PROJECTS_PATH

MAX_CONCURRENT_PROJECT_REQUESTS = 10
LOCK_TIMEOUT = 15


class Project(Model):
    """User project object."""

    __database__ = BaseCache.model_db
    __namespace__ = BaseCache.namespace

    created_at = DateTimeField()
    last_fetched_at = DateTimeField()

    project_id = TextField(primary_key=True, index=True)
    user_id = TextField(index=True)

    clone_depth = IntegerField()
    git_url = TextField(index=True)

    name = TextField()
    slug = TextField()
    fullname = TextField()
    description = TextField()
    email = TextField()
    owner = TextField()
    token = TextField()
    initialized = BooleanField()

    @property
    def abs_path(self):
        """Full path of cached project."""
        return CACHE_PROJECTS_PATH / self.user_id / self.owner / self.slug

    def read_lock(self, timeout: Optional[float] = None):
        """Shared read lock on the project."""
        timeout = timeout if timeout is not None else LOCK_TIMEOUT
        return portalocker.Lock(
            f"{self.abs_path}.lock", flags=portalocker.LOCK_SH | portalocker.LOCK_NB, timeout=timeout
        )

    def write_lock(self):
        """Exclusive write lock on the project."""
        return portalocker.Lock(f"{self.abs_path}.lock", flags=portalocker.LOCK_EX, timeout=LOCK_TIMEOUT)

    def concurrency_lock(self):
        """Lock to limit concurrent operations on a project.

        This serves as a "leaky bucket" type approach to prevent starvation with multiple
        concurrent requests.
        """
        return portalocker.BoundedSemaphore(
            MAX_CONCURRENT_PROJECT_REQUESTS, name=f"{self.name}_bounded_semaphore", directory=str(self.abs_path.parent)
        )

    @property
    def age(self):
        """Returns project's age in seconds."""
        # NOTE: `created_at` field is aligned to UTC timezone.
        return int((datetime.utcnow() - self.created_at).total_seconds())

    @property
    def fetch_age(self):
        """Returns project's fetch age in seconds."""
        return int((datetime.utcnow() - self.last_fetched_at).total_seconds())

    def exists(self):
        """Ensure a project exists on file system."""
        return self.abs_path.exists()

    def ttl_expired(self, ttl=None):
        """Check if project time to live has expired."""
        if not self.created_at:
            # If record does not contain created_at,
            # it means its an old record, and
            # we should mark it for deletion.
            return True

        # NOTE: time to live measured in seconds
        ttl = ttl or int(os.getenv("RENKU_SVC_CLEANUP_TTL_PROJECTS", 1800))
        return self.age >= ttl

    def purge(self):
        """Removes project from file system and cache."""
        shutil.rmtree(str(self.abs_path))
        self.delete()

    def is_locked(self, jobs):
        """Check if file locked by given jobs."""
        return bool(next((job for job in jobs if self.project_id in job.locked), False))



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Install and uninstall Git hooks."""

import stat
from pathlib import Path

from pydantic import validate_arguments

from renku.core.util.git import get_hook_path
from renku.domain_model.project_context import project_context

try:
    import importlib_resources  # type:ignore
except ImportError:
    import importlib.resources as importlib_resources  # type: ignore

HOOKS = ("pre-commit",)


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def install_githooks(force: bool, path: Path):
    """Install Git hooks."""
    warning_messages = []
    for hook in HOOKS:
        hook_path = get_hook_path(name=hook, path=path)
        if hook_path.exists():
            if not force:
                warning_messages.append(f"Hook already exists. Skipping {hook_path}")
                continue
            else:
                hook_path.unlink()

        # Make sure the hooks directory exists.
        hook_path.parent.mkdir(parents=True, exist_ok=True)
        hook_data = importlib_resources.files("renku.data").joinpath(f"{hook}.sh").read_bytes()
        Path(hook_path).write_bytes(hook_data)
        hook_path.chmod(hook_path.stat().st_mode | stat.S_IEXEC)

    return warning_messages


def uninstall_githooks():
    """Uninstall Git hooks."""
    for hook in HOOKS:
        hook_path = get_hook_path(name=hook, path=project_context.path)
        if hook_path.exists():
            hook_path.unlink()



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Dataset context managers."""

from pathlib import Path
from typing import Optional

from renku.core import errors
from renku.core.dataset.dataset import create_dataset
from renku.core.dataset.datasets_provenance import DatasetsProvenance
from renku.domain_model.dataset import Dataset
from renku.domain_model.project_context import project_context
from renku.domain_model.provenance.agent import Person


class DatasetContext:
    """Dataset context manager for metadata changes."""

    def __init__(
        self,
        name: str,
        create: Optional[bool] = False,
        commit_database: Optional[bool] = False,
        creator: Optional[Person] = None,
        datadir: Optional[Path] = None,
        storage: Optional[str] = None,
    ) -> None:
        self.name = name
        self.create = create
        self.commit_database = commit_database
        self.creator = creator
        self.dataset_provenance = DatasetsProvenance()
        self.dataset: Optional[Dataset] = None
        self.datadir: Optional[Path] = datadir
        self.storage = storage

    def __enter__(self):
        """Enter context."""
        self.dataset = self.dataset_provenance.get_by_name(name=self.name)
        if self.dataset is None:
            if not self.create:
                raise errors.DatasetNotFound(name=self.name)

            # NOTE: Don't update provenance when creating here because it will be updated later
            self.dataset = create_dataset(
                name=self.name, update_provenance=False, datadir=self.datadir, storage=self.storage
            )
        elif self.create:
            raise errors.DatasetExistsError(self.name)

        return self.dataset

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit context."""
        if exc_type:
            # TODO use a general clean-up strategy: https://github.com/SwissDataScienceCenter/renku-python/issues/736
            # NOTE: Re-raise exception
            return False

        if self.dataset and self.commit_database:
            self.datasets_provenance = DatasetsProvenance()
            self.datasets_provenance.add_or_update(self.dataset, creator=self.creator)
            project_context.database.commit()



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Initial migrations."""

import os
import shutil
import urllib
from pathlib import Path

from renku.core.constant import DEFAULT_DATA_DIR as DATA_DIR
from renku.core.constant import RENKU_HOME
from renku.core.migration.models.refs import LinkReference
from renku.core.migration.models.v3 import Collection, Dataset, Project, get_project_datasets
from renku.core.migration.models.v9 import generate_file_id, generate_label
from renku.core.migration.utils import (
    OLD_METADATA_PATH,
    generate_dataset_id,
    get_datasets_path,
    get_pre_0_3_4_datasets_metadata,
    is_using_temporary_datasets_path,
)
from renku.core.util.contexts import with_project_metadata
from renku.core.util.git import get_in_submodules
from renku.core.util.urls import url_to_string
from renku.domain_model.dataset import generate_default_name
from renku.domain_model.project_context import project_context


def migrate(migration_context):
    """Migration function."""
    _ensure_clean_lock()
    _do_not_track_lock_file()
    _migrate_datasets_pre_v0_3()
    _migrate_broken_dataset_paths(migration_context=migration_context)
    _fix_labels_and_ids(migration_context)
    _fix_dataset_urls()
    _migrate_dataset_and_files_project()


def _ensure_clean_lock():
    """Make sure Renku lock file is not part of repository."""
    if is_using_temporary_datasets_path():
        return

    lock_file = project_context.path / ".renku.lock"
    try:
        lock_file.unlink()
    except FileNotFoundError:
        pass


def _do_not_track_lock_file():
    """Add lock file to .gitignore if not already exists."""
    if is_using_temporary_datasets_path():
        return

    lock_file = ".renku.lock"
    gitignore = project_context.path / ".gitignore"
    if not gitignore.exists() or lock_file not in gitignore.read_text():
        gitignore.open("a").write(f"\n{lock_file}\n")


def _migrate_datasets_pre_v0_3():
    """Migrate datasets from Renku 0.3.x."""
    if is_using_temporary_datasets_path():
        return

    changed = False
    repository = project_context.repository

    for old_path in get_pre_0_3_4_datasets_metadata():
        changed = True
        name = str(old_path.parent.relative_to(project_context.path / DATA_DIR))

        dataset = Dataset.from_yaml(old_path)
        dataset.title = name
        dataset.name = generate_default_name(name)
        new_path = get_datasets_path() / dataset.identifier / OLD_METADATA_PATH
        new_path.parent.mkdir(parents=True, exist_ok=True)

        with with_project_metadata(read_only=True) as meta:
            for submodule in repository.submodules:
                if Path(submodule.url).name == meta.name:
                    repository.submodules.remove(submodule)

        for file_ in dataset.files:
            if not Path(file_.path).exists():
                expected_path = project_context.path / DATA_DIR / dataset.name / file_.path
                if expected_path.exists():
                    file_.path = expected_path.relative_to(project_context.path)

        dataset.to_yaml(new_path)

        Path(old_path).unlink()
        ref = LinkReference.create(name=f"datasets/{name}", force=True)
        ref.set_reference(new_path)

    if changed:
        project_path = project_context.metadata_path.joinpath(OLD_METADATA_PATH)
        project = Project.from_yaml(project_path)
        project.version = "3"
        project.to_yaml(project_path)

        repository.add(all=True)
        repository.commit("renku migrate: committing structural changes" + project_context.transaction_id)


def _migrate_broken_dataset_paths(migration_context):
    """Ensure all paths are using correct directory structure."""
    for dataset in get_project_datasets():
        if not dataset.name:
            dataset.name = generate_default_name(dataset.title, dataset.version)
        else:
            dataset.name = generate_default_name(dataset.name)

        expected_path = get_datasets_path() / dataset.identifier

        # migrate the refs
        if not is_using_temporary_datasets_path():
            ref = LinkReference.create(name=f"datasets/{dataset.name}", force=True)
            ref.set_reference(expected_path / OLD_METADATA_PATH)

        if not expected_path.exists():
            old_dataset_path = dataset.path
            if not is_using_temporary_datasets_path():
                expected_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.move(old_dataset_path, expected_path)
            else:
                expected_path.mkdir(parents=True, exist_ok=True)
                shutil.move(str(Path(old_dataset_path) / OLD_METADATA_PATH), expected_path)

        dataset.path = os.path.relpath(expected_path, project_context.path)

        if not is_using_temporary_datasets_path():
            base_path = project_context.path
        else:
            base_path = project_context.path / RENKU_HOME

        collections = [f for f in dataset.files if isinstance(f, Collection)]
        files = [f for f in dataset.files if not isinstance(f, Collection)]

        while collections:
            collection = collections.pop()
            for file in collection.members:
                if isinstance(file, Collection):
                    collections.append(file)
                else:
                    files.append(file)

        dataset.files = files

        for file in dataset.files:
            if _is_dir(migration_context=migration_context, path=file.path):
                continue
            if file.path.startswith(".."):
                file_absolute_path = os.path.abspath(get_datasets_path() / dataset.identifier / file.path)
                file.path = Path(file_absolute_path).relative_to(base_path)
            elif not _exists(migration_context=migration_context, path=file.path):
                file.path = (project_context.path / DATA_DIR / file.path).relative_to(project_context.path)

            file.name = os.path.basename(file.path)

        dataset.to_yaml(expected_path / "metadata.yml")


def _fix_labels_and_ids(migration_context):
    """Ensure files have correct label instantiation."""
    for dataset in get_project_datasets():
        dataset._id = generate_dataset_id(identifier=dataset.identifier)
        dataset._label = dataset.identifier

        for file in dataset.files:
            if not _exists(migration_context=migration_context, path=file.path):
                continue

            commit = _get_previous_commit(migration_context=migration_context, path=file.path)
            _, commit, _ = get_in_submodules(repository=project_context.repository, commit=commit, path=file.path)

            if not _is_file_id_valid(file._id, file.path, commit.hexsha):
                file._id = generate_file_id(hexsha=commit.hexsha, path=file.path)

            if not file._label or commit.hexsha not in file._label or file.path not in file._label:
                file._label = generate_label(file.path, commit.hexsha)

        dataset.to_yaml()


def _fix_dataset_urls():
    """Ensure dataset and its files have correct url format."""
    for dataset in get_project_datasets():
        dataset.url = dataset._id
        for file_ in dataset.files:
            if file_.url:
                file_.url = url_to_string(file_.url)

        dataset.to_yaml()


def _migrate_dataset_and_files_project():
    """Ensure dataset files have correct project."""
    project_path = project_context.metadata_path.joinpath(OLD_METADATA_PATH)
    project = Project.from_yaml(project_path)
    if not is_using_temporary_datasets_path():
        project.to_yaml(project_path)

    for dataset in get_project_datasets():
        dataset._project = project
        if not dataset.creators:
            dataset.creators = [project.creator]
        for file_ in dataset.files:
            file_._project = project

        dataset.to_yaml()


def _is_file_id_valid(id_, path, hexsha):
    if not id_ or not isinstance(id_, str) or not id_.startswith("https"):
        return False

    u = urllib.parse.urlparse(id_)
    return u.scheme and u.netloc and u.path.startswith("/blob") and hexsha in u.path and path in u.path


def _exists(migration_context, path):
    dmc = migration_context.dataset_migration_context
    if dmc:
        return dmc.exists(path)

    path = project_context.path / path
    return path.exists() or (path.is_symlink() and os.path.lexists(path))


def _is_dir(migration_context, path):
    dmc = migration_context.dataset_migration_context
    if dmc:
        return dmc.is_dir(path)

    return (project_context.path / path).is_dir()


def _get_previous_commit(migration_context, path):
    dmc = migration_context.dataset_migration_context
    if dmc:
        return dmc.get_previous_commit(path)
    return project_context.repository.get_previous_commit(path, revision="HEAD")



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service cache serializers."""
import time
import uuid
from urllib.parse import urlparse

from marshmallow import Schema, ValidationError, fields, post_load, pre_load, validates_schema
from marshmallow_oneofschema import OneOfSchema
from werkzeug.utils import secure_filename

from renku.core import errors
from renku.core.util.os import normalize_to_ascii
from renku.domain_model.git import GitURL
from renku.ui.service.config import PROJECT_CLONE_DEPTH_DEFAULT
from renku.ui.service.serializers.common import (
    ArchiveSchema,
    AsyncSchema,
    ErrorResponse,
    FileDetailsSchema,
    LocalRepositorySchema,
    RemoteRepositorySchema,
    RenkuSyncSchema,
)
from renku.ui.service.serializers.rpc import JsonRPCResponse


def extract_file(request):
    """Extract file from Flask request.

    Raises:
        `ValidationError`: If file data or filename is missing in request.
    """
    files = request.files
    if "file" not in files:
        raise ValidationError("missing key: file")

    file = files["file"]
    if file and not file.filename:
        raise ValidationError(f"wrong filename: {file.filename}")

    if file:
        file.filename = secure_filename(file.filename)
        return file


class FileUploadRequest(ArchiveSchema):
    """Request schema for file upload."""

    override_existing = fields.Boolean(
        load_default=False,
        metadata={"description": "Overried files. Useful when extracting from archives."},
    )

    chunked_id = fields.String(data_key="dzuuid", load_default=None, metadata={"description": "Dropzone upload id."})
    chunk_index = fields.Integer(
        data_key="dzchunkindex", load_default=None, metadata={"description": "Dropzone chunk index."}
    )
    chunk_count = fields.Integer(
        data_key="dztotalchunkcount", load_default=None, metadata={"description": "Dropzone total chunk count."}
    )
    chunk_size = fields.Integer(
        data_key="dzchunksize", load_default=None, metadata={"description": "Dropzone chunk size."}
    )
    chunk_byte_offset = fields.Integer(
        data_key="dzchunkbyteoffset", load_default=None, metadata={"description": "Dropzone chunk offset."}
    )
    chunked_content_type = fields.String(
        load_default=None, metadata={"description": "Content type of file file for chunked uploads."}
    )
    total_size = fields.Integer(
        data_key="dztotalfilesize", load_default=None, metadata={"description": "Dropzone total file size."}
    )

    @validates_schema
    def validate_requires(self, data, **kwargs):
        """Validate chunk upload data."""
        if data.get("dzuuid") is not None and ("dzchunkindex" not in data or "dztotalchunkcount" not in data):
            raise ValidationError("'dzchunkindex' and 'dztotalchunkcount' are required when 'dzuuid' is set.")


class FileUploadResponse(Schema):
    """Response schema for file upload."""

    files = fields.List(fields.Nested(FileDetailsSchema))


class FileUploadResponseRPC(JsonRPCResponse):
    """RPC response schema for file upload response."""

    result = fields.Nested(FileUploadResponse)


class FileChunksDeleteRequest(Schema):
    """Request schema for deleting uploaded chunks."""

    chunked_id = fields.String(data_key="dzuuid", load_default=None, metadata={"description": "Dropzone upload id."})


class FileChunksDeleteResponseRPC(JsonRPCResponse):
    """RPC response schema for file upload response."""

    result = fields.String()


class FileListResponse(Schema):
    """Response schema for files listing."""

    files = fields.List(fields.Nested(FileDetailsSchema), required=True)


class FileListResponseRPC(JsonRPCResponse):
    """RPC response schema for files listing."""

    result = fields.Nested(FileListResponse)


class RepositoryCloneRequest(RemoteRepositorySchema):
    """Request schema for repository clone."""

    depth = fields.Integer(metadata={"description": "Git fetch depth"}, load_default=PROJECT_CLONE_DEPTH_DEFAULT)


class ProjectCloneContext(RepositoryCloneRequest):
    """Context schema for project clone."""

    # measured in ms
    timestamp = fields.Integer(load_default=time.time() * 1e3)

    # user data
    name = fields.String()
    slug = fields.String()
    fullname = fields.String()
    email = fields.String()
    owner = fields.String()
    token = fields.String()

    @pre_load()
    def set_missing_id(self, data, **kwargs):
        """Set project_id when missing."""
        if not data.get("project_id"):
            data["project_id"] = uuid.uuid4().hex

        return data

    @pre_load()
    def set_owner_name(self, data, **kwargs):
        """Set owner and name fields."""
        try:
            git_url = GitURL.parse(data["git_url"])
        except UnicodeError as e:
            raise ValidationError("`git_url` contains unsupported characters") from e
        except errors.InvalidGitURL as e:
            raise ValidationError("Invalid `git_url`") from e

        if git_url.owner is None:
            raise ValidationError("Invalid `git_url`")
        data["owner"] = git_url.owner

        if git_url.name is None:
            raise ValidationError("Invalid `git_url`")
        data["name"] = git_url.name
        data["slug"] = normalize_to_ascii(data["name"])

        return data

    def format_url(self, data):
        """Format url with auth."""
        git_url = urlparse(data["git_url"])

        url = "oauth2:{}@{}".format(data["token"], git_url.netloc)
        return git_url._replace(netloc=url).geturl()

    @post_load
    def finalize_data(self, data, **kwargs):
        """Finalize data."""
        data["url_with_auth"] = self.format_url(data)

        if not data["depth"]:
            # NOTE: In case of `depth=None` or `depth=0` we set to default depth.
            data["depth"] = PROJECT_CLONE_DEPTH_DEFAULT

        try:
            depth = int(data["depth"])

            if depth < 0:
                # NOTE: In case of `depth<0` we remove the depth limit.
                data["depth"] = None

        except ValueError:
            data["depth"] = PROJECT_CLONE_DEPTH_DEFAULT

        return data


class ProjectCloneResponse(Schema):
    """Response schema for project clone."""

    project_id = fields.String(required=True)
    git_url = fields.String(required=True)
    initialized = fields.Boolean(dump_default=False)


class ProjectCloneResponseRPC(JsonRPCResponse):
    """RPC response schema for project clone response."""

    result = fields.Nested(ProjectCloneResponse)


class ProjectListResponse(Schema):
    """Response schema for project listing."""

    projects = fields.List(fields.Nested(ProjectCloneResponse), required=True)


class ProjectListResponseRPC(JsonRPCResponse):
    """RPC response schema for project listing."""

    result = fields.Nested(ProjectListResponse)


class ProjectMigrateRequest(AsyncSchema, LocalRepositorySchema, RemoteRepositorySchema):
    """Request schema for project migrate."""

    force_template_update = fields.Boolean(dump_default=False)
    skip_template_update = fields.Boolean(dump_default=False)
    skip_docker_update = fields.Boolean(dump_default=False)
    skip_migrations = fields.Boolean(dump_default=False)


class ProjectMigrateResponse(RenkuSyncSchema):
    """Response schema for project migrate."""

    was_migrated = fields.Boolean()
    template_migrated = fields.Boolean()
    docker_migrated = fields.Boolean()
    messages = fields.List(fields.String)
    warnings = fields.List(fields.String)
    errors = fields.List(fields.String)


class ProjectMigrateResponseRPC(JsonRPCResponse):
    """RPC response schema for project migrate."""

    result = fields.Nested(ProjectMigrateResponse)


class ProjectMigrationCheckRequest(LocalRepositorySchema, RemoteRepositorySchema):
    """Request schema for project migration check."""


class ProjectCompatibilityResponseDetail(Schema):
    """Response schema outlining service compatibility for migrations check."""

    project_metadata_version = fields.String(
        metadata={"description": "Current version of the Renku metadata in the project."}
    )
    current_metadata_version = fields.String(
        metadata={"description": "Highest metadata version supported by this service."}
    )
    migration_required = fields.Boolean(
        metadata={"description": "Whether or not a metadata migration is required to be compatible with this service."}
    )
    fixes_available = fields.Boolean(
        metadata={
            "description": "Whether automated fixes of metadata (beyond those done during migration) are available."
        }
    )
    issues_found = fields.List(fields.Str, metadata={"description": "Metadata issues found on project."})


class ProjectCompatibilityResponse(OneOfSchema):
    """Combined schema of DockerfileStatusResponseDetail or Exception."""

    type_schemas = {"detail": ProjectCompatibilityResponseDetail, "error": ErrorResponse}

    def get_obj_type(self, obj):
        """Get type from object."""
        from renku.command.migrate import CoreStatusResult

        if isinstance(obj, CoreStatusResult) or (isinstance(obj, dict) and "userMessage" not in obj):
            return "detail"

        return "error"


class DockerfileStatusResponseDetail(Schema):
    """Response schema outlining dockerfile status for migrations check."""

    newer_renku_available = fields.Boolean(
        metadata={
            "description": "Whether the version of Renku in this service is newer than the one in the Dockerfile."
        }
    )
    automated_dockerfile_update = fields.Boolean(
        metadata={"description": "Whether or not the Dockerfile supports automated Renku version updates."}
    )
    latest_renku_version = fields.String(
        metadata={"description": "The current version of Renku available in this service."}
    )
    dockerfile_renku_version = fields.String(metadata={"description": "Version of Renku specified in the Dockerfile."})


class DockerfileStatusResponse(OneOfSchema):
    """Combined schema of DockerfileStatusResponseDetail or Exception."""

    type_schemas = {"detail": DockerfileStatusResponseDetail, "error": ErrorResponse}

    def get_obj_type(self, obj):
        """Get type from object."""
        from renku.command.migrate import DockerfileStatusResult

        if isinstance(obj, DockerfileStatusResult) or (isinstance(obj, dict) and "userMessage" not in obj):
            return "detail"

        return "error"


class TemplateStatusResponseDetail(Schema):
    """Response schema outlining template status for migrations check."""

    automated_template_update = fields.Boolean(
        metadata={"description": "Whether or not the project template explicitly supports automated updates."}
    )
    newer_template_available = fields.Boolean(
        metadata={
            "description": "Whether or not the current version of the project template differs from the "
            "one used in the project."
        }
    )
    template_source = fields.String(
        metadata={
            "description": "Source of the template repository, "
            "either a Git URL or 'renku' if an embedded template was used."
        }
    )
    template_ref = fields.String(
        metadata={
            "description": "The branch/tag/commit from the template_source repository "
            "that was used to create this project."
        }
    )
    template_id = fields.String(metadata={"description": "The id of the template in the template repository."})

    project_template_version = fields.String(
        allow_none=True, metadata={"description": "The version of the template last used in the user's project."}
    )
    latest_template_version = fields.String(
        allow_none=True, metadata={"description": "The current version of the template in the template repository."}
    )

    ssh_supported = fields.Boolean(
        metadata={"description": "Whether this project supports ssh connections to sessions or not."}
    )


class TemplateStatusResponse(OneOfSchema):
    """Combined schema of TemplateStatusResponseDetail or Exception."""

    type_schemas = {"detail": TemplateStatusResponseDetail, "error": ErrorResponse}

    def get_obj_type(self, obj):
        """Get type from object."""
        from renku.command.migrate import TemplateStatusResult

        if isinstance(obj, TemplateStatusResult) or (isinstance(obj, dict) and "userMessage" not in obj):
            return "detail"

        return "error"


class ProjectMigrationCheckResponse(Schema):
    """Response schema for project migration check."""

    project_supported = fields.Boolean(
        metadata={
            "description": "Determines whether this project is a Renku project that is supported by the version "
            "running on this service (not made with a newer version)."
        }
    )
    core_renku_version = fields.String(metadata={"description": "Version of Renku running in this service."})
    project_renku_version = fields.String(metadata={"description": "Version of Renku last used to change the project."})

    core_compatibility_status = fields.Nested(
        ProjectCompatibilityResponse,
        metadata={"description": "Fields detailing the compatibility of the project with this core service version."},
    )
    dockerfile_renku_status = fields.Nested(
        DockerfileStatusResponse,
        metadata={"description": "Fields detailing the status of the Dockerfile in the project."},
    )
    template_status = fields.Nested(
        TemplateStatusResponse,
        metadata={"description": "Fields detailing the status of the project template used by this project."},
    )


class ProjectMigrationCheckResponseRPC(JsonRPCResponse):
    """RPC response schema for project migration check."""

    result = fields.Nested(ProjectMigrationCheckResponse)



#
# Copyright 2020 - Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Renku service migrate project controller."""
from renku.core.util.contexts import renku_project_context
from renku.ui.service.cache.models.job import Job
from renku.ui.service.config import MESSAGE_PREFIX, PROJECT_CLONE_NO_DEPTH
from renku.ui.service.controllers.api.abstract import ServiceCtrl
from renku.ui.service.controllers.api.mixins import RenkuOpSyncMixin
from renku.ui.service.logger import worker_log
from renku.ui.service.serializers.cache import ProjectMigrateRequest, ProjectMigrateResponseRPC
from renku.ui.service.utils.callback import ServiceCallback
from renku.ui.service.views import result_response
from renku.version import __version__


def execute_migration(
    project_path, force_template_update, skip_template_update, skip_docker_update, skip_migrations, commit_message
):
    """Execute project migrations."""
    from renku.command.migrate import migrate_project_command

    worker_log.debug(f"migrating {project_path}")

    communicator = ServiceCallback()

    with renku_project_context(project_path):
        result = (
            migrate_project_command()
            .with_commit(message=commit_message)
            .with_communicator(communicator)
            .build()
            .execute(
                force_template_update=force_template_update,
                skip_template_update=skip_template_update,
                skip_docker_update=skip_docker_update,
                skip_migrations=skip_migrations,
            )
        )

        was_migrated, template_migrated, docker_migrated = result.output

    worker_log.debug(f"migration finished - was_migrated={was_migrated}")
    return (
        was_migrated,
        template_migrated,
        docker_migrated,
        communicator.messages,
        communicator.warnings,
        communicator.errors,
    )


class MigrateProjectCtrl(ServiceCtrl, RenkuOpSyncMixin):
    """Controller for migrating project endpoint."""

    REQUEST_SERIALIZER = ProjectMigrateRequest()
    RESPONSE_SERIALIZER = ProjectMigrateResponseRPC()

    def __init__(self, cache, user_data, request_data, migrate_project=False, skip_lock=False):
        """Construct controller."""
        self.ctx = MigrateProjectCtrl.REQUEST_SERIALIZER.load(request_data)

        self.force_template_update = self.ctx.get("force_template_update", False)
        self.skip_template_update = self.ctx.get("skip_template_update", False)
        self.skip_docker_update = self.ctx.get("skip_docker_update", False)
        self.skip_migrations = self.ctx.get("skip_migrations", False)
        self.commit_message = f"{MESSAGE_PREFIX} migrate to latest version for renku {__version__}"

        super().__init__(
            cache,
            user_data,
            request_data,
            migrate_project=migrate_project,
            skip_lock=skip_lock,
            clone_depth=PROJECT_CLONE_NO_DEPTH,
        )

    @property
    def context(self):
        """Controller operation context."""
        return self.ctx

    def renku_op(self):
        """Renku operation for the controller."""
        return execute_migration(
            self.project_path,
            self.force_template_update,
            self.skip_template_update,
            self.skip_docker_update,
            self.skip_migrations,
            self.commit_message,
        )

    def to_response(self):
        """Execute controller flow and serialize to service response."""
        op_result, remote_branch = self.execute_and_sync()
        if isinstance(op_result, Job):
            return result_response(MigrateProjectCtrl.JOB_RESPONSE_SERIALIZER, op_result)

        was_migrated, template_migrated, docker_migrated, messages, warnings, errors = op_result

        response = {
            "messages": messages,
            "warnings": warnings,
            "errors": errors,
            "was_migrated": was_migrated,
            "template_migrated": template_migrated,
            "docker_migrated": docker_migrated,
            "remote_branch": remote_branch,
        }

        return result_response(self.RESPONSE_SERIALIZER, response)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Project view model."""

import json
from datetime import datetime
from typing import List, Optional

from renku.core.config import get_value
from renku.domain_model.project import Project
from renku.domain_model.provenance.agent import Person


class ProjectViewModel:
    """A view model for a ``Project``."""

    def __init__(
        self,
        id: str,
        name: Optional[str],
        creator: Person,
        created: datetime,
        description: Optional[str],
        agent: Optional[str],
        annotations: Optional[str],
        template_info: str,
        keywords: Optional[List[str]],
        ssh_supported: bool = False,
    ):
        self.id = id
        self.name = name
        self.creator = creator
        self.creator_str = creator.full_identity
        self.created = created
        self.created_str = created.isoformat()
        self.description = description
        self.agent = agent
        self.annotations = annotations
        self.template_info = template_info
        self.keywords = keywords
        self.keywords_str = ", ".join(keywords) if keywords else ""
        self.ssh_supported = ssh_supported

    @classmethod
    def from_project(cls, project: Project):
        """Create view model from ``Project``.

        Args:
            project(Project): The project to convert.

        Returns:
            View model for project.
        """
        template_info = ""

        if project.template_metadata.template_source:
            if project.template_metadata.template_source == "renku":
                template_info = (
                    f"{project.template_metadata.template_id} ({project.template_metadata.template_version})"
                )
            else:
                template_info = (
                    f"{project.template_metadata.template_source}@"
                    f"{project.template_metadata.template_ref}: {project.template_metadata.template_id}"
                )

        return cls(
            id=project.id,
            name=project.name,
            creator=project.creator,
            created=project.date_created,
            description=project.description,
            agent=project.agent_version,
            annotations=json.dumps([{"id": a.id, "body": a.body, "source": a.source} for a in project.annotations])
            if project.annotations
            else None,
            template_info=template_info,
            keywords=project.keywords,
            ssh_supported=get_value("renku", "ssh_supported") == "true" or project.template_metadata.ssh_supported,
        )



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Activity graph view model."""

import abc
import math
from collections import defaultdict
from copy import deepcopy
from dataclasses import dataclass
from functools import partial
from typing import Any, Callable, Dict, Generator, NamedTuple, Optional, Tuple

from click import style


class Point(NamedTuple):
    """A point with coordinates for rendering."""

    x: int
    y: int


class Size(NamedTuple):
    """A size for a two dimensional matrix."""

    width: int
    height: int


class EdgeSet(NamedTuple):
    """A collection of edge characters for rendering boxes."""

    top_left: str
    bottom_left: str
    top_right: str
    bottom_right: str


@dataclass
class CharacterSet:
    """A rendering character set for rendering Ascii content."""

    edge: EdgeSet
    horizontal: str
    vertical: str

    def style(self, transform: Callable[[str], str]) -> None:
        """Apply a style to this character set (like bold).

        Args:
            transform(Callable[[str], str]): Function that applies a style to a string.
        """
        self.horizontal = transform(self.horizontal)
        self.vertical = transform(self.vertical)

        self.edge = EdgeSet(
            transform(self.edge.top_left),
            transform(self.edge.bottom_left),
            transform(self.edge.top_right),
            transform(self.edge.bottom_right),
        )


MAX_NODE_LENGTH = 40


def line_points(start: Point, end: Point) -> Generator[Point, None, None]:
    """Return all (discrete) points of a line from start to end.

    Args:
        start(Point): Starting point of the line.
        end(Point): End point of the line.
    Returns:
        Generator[Point, None, None]: All discrete points part of the line.
    """
    x_extent = end.x - start.x
    y_extent = end.y - start.y

    length = max(abs(x_extent), abs(y_extent)) + 1
    assert length > 0
    for i in range(length):
        position = i / length
        yield Point(math.ceil(start.x + position * x_extent), math.ceil(start.y + position * y_extent))


class TextMatrix:
    """A two dimensional matrix of strings.

    Indices need to be positive. To support handling negative indices, you can set x/y-offset to the lowest
    negative index you want to pass and indices will be shifted by this amount internally.

    Args:
        size(Tuple[int,int]): The width and height of the matrix.
        x_offset(int): Offset to apply to x indices.
        y_offset(int): Offset to apply to y indices.
    """

    def __init__(self, size: Size, x_offset: int = 0, y_offset: int = 0) -> None:
        assert size.width > 0
        assert size.height > 0
        self._size = size
        self.x_offset = x_offset
        self.y_offset = y_offset
        self._content = []

        for _ in range(self.size.height):
            row = []
            for _ in range(self.size.width):
                row.append(" ")
            self._content.append(row)

    @property
    def size(self) -> Size:
        """The size of the matrix."""
        return self._size

    def __getitem__(self, index: Tuple[int, int]) -> str:
        """Get element at width,height position."""
        pos = (index[0] + self.x_offset, index[1] + self.y_offset)
        assert 0 <= pos[0] < self.size.width
        assert 0 <= pos[1] < self.size.height
        return self._content[pos[1]][pos[0]]

    def __setitem__(self, index: Tuple[int, int], value: str) -> None:
        """Set element at width,height position."""
        pos = (index[0] + self.x_offset, index[1] + self.y_offset)
        assert 0 <= pos[0] < self.size.width
        assert 0 <= pos[1] < self.size.height
        self._content[pos[1]][pos[0]] = value

    def __str__(self) -> str:
        return "\n".join(["".join(row) for row in self._content])

    def draw_line(self, start: Point, end: Point, value: str) -> None:
        """Fill a line from start to end with value.

        Args:
            start(Point): The starting point.
            end(Point): The end point.
            value(str): The value (character) to use for representing the line.
        """
        for point in line_points(start, end):
            self[point.x, point.y] = value

    def draw(self, start: Point, value: "TextMatrix") -> None:
        """Write content from value with start being upper-left corner.

        Args:
            start(Point): The upper-left corner for where to place the content.
            value(TextMatrix): The content text matrix to place into this one.
        """
        for y in range(start.y, start.y + value.size.height):
            for x in range(start.x, start.x + value.size.width):
                self[x, y] = value[x - start.x, y - start.y]


class Shape(abc.ABC):
    """Basic shape class that all shapes inherit from."""

    def draw(self, matrix: TextMatrix, color: bool = True, ascii=False) -> None:
        """Draw self onto a text matrix..

        Args:
            matrix(TextMatrix): The text matrix to draw on.
            color(bool, optional): Whether or not to render in color (Default value = True).
            ascii:  Whether to use ascii characters only or with UTF8 (Default value = False).
        """
        raise NotImplementedError()


class RectangleShape(Shape):
    """A rectangle shape."""

    ASCII_CHARACTERS = CharacterSet(edge=EdgeSet("+", "+", "+", "+"), horizontal="-", vertical="|")

    UNICODE_CHARACTERS = CharacterSet(edge=EdgeSet("┌", "└", "┐", "┘"), horizontal="─", vertical="│")

    UNICODE_CHARACTERS_DOUBLE = CharacterSet(edge=EdgeSet("╔", "╚", "╗", "╝"), horizontal="═", vertical="║")

    def __init__(self, start: Point, end: Point, double_border=False, color: Optional[str] = None):
        self.start = start
        self.end = end
        self.double_border = double_border
        self.color = color

    def draw(self, matrix: TextMatrix, color: bool = True, ascii=False) -> None:
        """Draw self onto a text matrix..

        Args:
            matrix(TextMatrix): The text matrix to draw on.
            color(bool, optional): Whether or not to render in color (Default value = True).
            ascii:  Whether to use ascii characters only or with UTF8 (Default value = False).
        """
        if not ascii and self.double_border:
            characters = deepcopy(self.UNICODE_CHARACTERS_DOUBLE)
        else:
            characters = deepcopy(self.ASCII_CHARACTERS if ascii else self.UNICODE_CHARACTERS)

        if color and self.color:
            # NOTE: Add color to border characters
            characters.style(partial(style, fg=self.color))
        # horizontal lines
        matrix.draw_line(
            Point(self.start.x + 1, self.start.y), Point(self.end.x - 1, self.start.y), characters.horizontal
        )
        matrix.draw_line(Point(self.start.x + 1, self.end.y), Point(self.end.x - 1, self.end.y), characters.horizontal)

        # vertical lines
        matrix.draw_line(Point(self.start.x, self.start.y + 1), Point(self.start.x, self.end.y), characters.vertical)
        matrix.draw_line(Point(self.end.x, self.start.y + 1), Point(self.end.x, self.end.y), characters.vertical)

        # set corners
        matrix[self.start.x, self.start.y] = characters.edge.top_left
        matrix[self.start.x, self.end.y] = characters.edge.bottom_left
        matrix[self.end.x, self.start.y] = characters.edge.top_right
        matrix[self.end.x, self.end.y] = characters.edge.bottom_right

        # fill with whitespace to force overwriting underlying text
        content = TextMatrix(Size(self.end.x - self.start.x - 1, self.end.y - self.start.y - 1))
        matrix.draw(Point(self.start.x + 1, self.start.y + 1), content)

    @property
    def extent(self) -> Tuple[Point, Point]:
        """The extent of this shape.

        Returns:
            Bounds of this shape.
        """
        return self.start, self.end


class TextShape(Shape):
    """A text object."""

    def __init__(self, text: str, point: Point, bold: bool = False, color: Optional[str] = None):
        self.point = point
        self.text = text.splitlines()
        self.bold = bold
        self.color = color

    def draw(self, matrix: TextMatrix, color: bool = True, ascii=False) -> None:
        """Draw self onto a text matrix..

        Args:
            matrix(TextMatrix): The text matrix to draw on.
            color(bool, optional): Whether or not to render in color (Default value = True).
            ascii:  Whether to use ascii characters only or with UTF8 (Default value = False).
        """

        for y, line in enumerate(self.text):
            for x, char in enumerate(line):
                kwargs: Dict[str, Any] = dict()
                if self.bold:
                    kwargs["bold"] = True
                if color and self.color:
                    kwargs["fg"] = self.color

                if kwargs:
                    char = style(char, **kwargs)

                matrix[self.point.x + x, self.point.y + y] = char

    @property
    def extent(self) -> Tuple[Point, Point]:
        """The extent of this shape.

        Returns:
            Bounds of this shape.
        """
        max_line_len = max(len(line) for line in self.text)
        num_lines = len(self.text)
        return (self.point, Point(self.point.x + max_line_len - 1, self.point.y + num_lines - 1))


class NodeShape(Shape):
    """An activity node shape."""

    def __init__(self, text: str, point: Point, double_border=False, color: Optional[str] = None):
        self.point = point
        self.text_shape = TextShape(text, Point(self.point.x + 1, self.point.y + 1), bold=double_border, color=color)
        text_extent = self.text_shape.extent
        self.box_shape = RectangleShape(
            self.point,
            Point(text_extent[1].x + 1, text_extent[1].y + 1),
            double_border=double_border,
            color=color,
        )

    def draw(self, matrix: TextMatrix, color: bool = True, ascii=False) -> None:
        """Draw self onto a text matrix..

        Args:
            matrix(TextMatrix): The text matrix to draw on.
            color(bool, optional): Whether or not to render in color (Default value = True).
            ascii:  Whether to use ascii characters only or with UTF8 (Default value = False).
        """
        self.actual_extent = self.extent
        self.box_shape.draw(matrix, color=color, ascii=ascii)
        self.text_shape.draw(matrix, color=color, ascii=ascii)

    @property
    def extent(self) -> Tuple[Point, Point]:
        """The extent of this shape.

        Returns:
            Bounds of this shape.
        """
        box_extent = self.box_shape.extent
        return Point(box_extent[0].x, box_extent[0].y), Point(
            box_extent[1].x,
            box_extent[1].y,
        )


class EdgeShape(Shape):
    """An edge between two activities."""

    COLORS = ["red", "green", "yellow", "blue", "magenta", "cyan"]
    CURRENT_COLOR = 0

    def __init__(self, start: Point, end: Point, color: str):
        self.start = Point(round(start.x), round(start.y))
        self.end = Point(round(end.x), round(end.y))
        self.color = color

    @staticmethod
    def next_color() -> str:
        """Get the next color in the color rotation.

        Returns:
            Next color string to use.
        """
        EdgeShape.CURRENT_COLOR = (EdgeShape.CURRENT_COLOR + 1) % len(EdgeShape.COLORS)
        return EdgeShape.COLORS[EdgeShape.CURRENT_COLOR]

    def intersects_with(self, other_edge: "EdgeShape") -> bool:
        """Checks whether this edge intersects with other edges.

        Args:
            other_edge("EdgeShape"): Edge to check intersection with.

        Returns:
            bool: True if this edge intersects ``other_edge``, False otherwise.
        """
        return not set(line_points(self.start, self.end)).intersection(set(line_points(self.start, self.end)))

    def draw(self, matrix: TextMatrix, color: bool = True, ascii=False) -> None:
        """Draw self onto a text matrix..

        Args:
            matrix(TextMatrix): The text matrix to draw on.
            color(bool, optional): Whether or not to render in color (Default value = True).
            ascii:  Whether to use ascii characters only or with UTF8 (Default value = False).
        """
        char = "*"

        if color:
            char = style(char, fg=self.color)

        matrix.draw_line(self.start, self.end, char)

    @property
    def extent(self) -> Tuple[Point, Point]:
        """The extent of this shape.

        Returns:
            Bounds of this shape.
        """
        return Point(min(self.start.x, self.end.x), min(self.start.y, self.end.y)), Point(
            max(self.start.x, self.end.x), max(self.start.y, self.end.y)
        )


class TextCanvas:
    """A canvas that can draw shapes to text."""

    def __init__(self):
        self.shapes = defaultdict(list)
        self._canvas = None

    def add_shape(self, shape: Shape, layer: int = 0) -> None:
        """Add a shape to the canvas.

        Args:
            shape(Shape): The shape to add.
            layer(int, optional): The layer to add the shape to (Default value = 0).
        """
        self.shapes[layer].append(shape)

    def get_coordinates(self, point: Point) -> Point:
        """Get actual, transformed coordinates of a node.

        Args:
            point(Point): The point to get coordinates for.

        Returns:
            Point: Coordinates in parent coordinate system.
        """
        return Point(point.x + self.x_offset, point.y)

    def render(self, color: bool = True, ascii=False):
        """Render contained shapes onto canvas.

        Args:
            color(bool, optional): Whether or not to render in color (Default value = True).
            ascii:  Whether to use ascii characters only or with UTF8 (Default value = False).
        """
        extent = (Point(2**64, 2**64), Point(-(2**64), -(2**64)))

        layers = sorted(self.shapes.keys())

        for layer in layers:
            for shape in self.shapes[layer]:
                shape_extent = shape.extent
                extent = (
                    Point(
                        math.floor(min(extent[0].x, shape_extent[0].x)), math.floor(min(extent[0].y, shape_extent[0].y))
                    ),
                    Point(
                        math.ceil(max(extent[1].x, shape_extent[1].x)), math.ceil(max(extent[1].y, shape_extent[1].y))
                    ),
                )
        self.x_offset = -extent[0].x
        size = Size(extent[1].x - extent[0].x + 30, extent[1].y - extent[0].y + 2)
        self._canvas = TextMatrix(size, x_offset=self.x_offset, y_offset=-extent[0].y)

        for layer in layers:
            for shape in self.shapes[layer]:
                shape.draw(self._canvas, color=color, ascii=ascii)

    @property
    def text(self) -> str:
        """Get the text of the canvas.

        Returns:
            String containing rendered canvas.
        """
        if self._canvas is None:
            raise ValueError("Call render() before getting text.")

        return str(self._canvas)



# Copyright Swiss Data Science Center (SDSC). A partnership between
# École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Log view model."""

from dataclasses import asdict, dataclass
from datetime import datetime
from enum import Enum
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple

import inject

if TYPE_CHECKING:
    from renku.command.view_model.plan import PlanViewModel
    from renku.domain_model.dataset import Dataset
    from renku.domain_model.provenance.activity import Activity


def tabular(data, columns) -> str:
    """Tabular output.

    Args:
        data: Input data.
        columns: Columns to show.

    Returns:
        str: data in tabular form.
    """
    from renku.command.format.tabulate import tabulate

    return tabulate(data, columns, columns_mapping=LOG_COLUMNS, reverse=True)


def json(data, columns) -> str:
    """JSON output.

    Args:
        data: Input data.
        columns: Not used.

    Returns:
        str: Data in JSON format.
    """
    import json

    data = sorted(data, key=lambda x: x.date)
    return json.dumps([d.to_dict() for d in data], indent=2)


class LogType(str, Enum):
    """Enum of different types of Log entries."""

    ACTIVITY = "Run"
    DATASET = "Dataset"


@dataclass
class ActivityDetailsViewModel:
    """View Model for Activity details."""

    start_time: str
    end_time: str
    renku_version: Optional[str] = None
    user: Optional[str] = None
    inputs: Optional[List[Tuple[str, str]]] = None
    outputs: Optional[List[Tuple[str, str]]] = None
    parameters: Optional[List[Tuple[str, str]]] = None


@dataclass
class DatasetChangeDetailsViewModel:
    """View model for detailed changes in a dataset."""

    created: bool = False
    imported: bool = False
    deleted: bool = False
    migrated: bool = False
    modified: bool = False
    files_added: Optional[List[str]] = None
    files_removed: Optional[List[str]] = None
    title_changed: Optional[str] = None
    description_changed: Optional[str] = None
    creators_added: Optional[List[str]] = None
    creators_removed: Optional[List[str]] = None
    keywords_added: Optional[List[str]] = None
    keywords_removed: Optional[List[str]] = None
    images_changed_to: Optional[List[str]] = None
    source: Optional[str] = None


class LogViewModel:
    """ViewModel for renku log entries."""

    def __init__(
        self,
        id: str,
        date: datetime,
        description: str,
        agents: List[str],
    ):
        self.id = id
        self.date = date
        self.description = description
        self.agents = agents

    def to_dict(self) -> Dict[str, Any]:
        """Return a dict representation of this view model."""
        raise NotImplementedError()

    @classmethod
    def from_activity(cls, activity: "Activity") -> "ActivityLogViewModel":
        """Create a log entry from an activity.

        Args:
            activity("Activity"): Activity to create log entry from.

        Returns:
            Log entry for activity.
        """
        from renku.command.view_model.plan import PlanViewModel
        from renku.domain_model.provenance.agent import Person, SoftwareAgent

        plan = activity.plan_with_values

        details = ActivityDetailsViewModel(
            start_time=activity.started_at_time.isoformat(), end_time=activity.ended_at_time.isoformat()
        )

        user = next((a for a in activity.agents if isinstance(a, Person)), None)
        renku_user = next((a for a in activity.agents if isinstance(a, SoftwareAgent)), None)

        if user:
            details.user = user.full_identity

        if renku_user:
            details.renku_version = renku_user.name

        command = " ".join(activity.plan_with_values.to_argv())

        if plan.inputs:
            details.inputs = [(input.name, input.actual_value) for input in plan.inputs]

        if plan.outputs:
            details.outputs = [(output.name, output.actual_value) for output in plan.outputs]

        if plan.parameters:
            details.parameters = [(parameter.name, parameter.actual_value) for parameter in plan.parameters]

        return ActivityLogViewModel(
            id=activity.id,
            date=activity.ended_at_time or activity.started_at_time or datetime.utcfromtimestamp(0),
            description=command,
            details=details,
            agents=[a.full_identity for a in activity.agents],
            plan=PlanViewModel.from_plan(activity.plan_with_values),
        )

    @classmethod
    def from_dataset(cls, dataset: "Dataset") -> "DatasetLogViewModel":
        """Create a log entry from an activity.

        Args:
            dataset("Dataset"): Dataset to create log entry for.

        Returns:
            Log entry for dataset.
        """
        from renku.core.interface.dataset_gateway import IDatasetGateway

        dataset_gateway = inject.instance(IDatasetGateway)

        descriptions = [f"Dataset '{dataset.name}': "]
        details = DatasetChangeDetailsViewModel()

        if not dataset.derived_from and not dataset.same_as:
            descriptions.append("created")
            details.created = True
        elif dataset.same_as:
            descriptions.append("imported")
            details.imported = True
            details.source = dataset.same_as.value
        elif dataset.derived_from and dataset.date_removed:
            descriptions.append("deleted")
            details.deleted = True

        previous_dataset: Optional[Dataset] = None

        if dataset.is_derivation():
            previous_dataset = dataset_gateway.get_by_id(dataset.derived_from.value)  # type: ignore

        current_files = {f for f in dataset.dataset_files if not f.date_removed}
        previous_files = set()

        if previous_dataset:
            previous_files = {f for f in previous_dataset.dataset_files if not f.date_removed}

        if (
            previous_files
            and {f.id for f in current_files}.difference({f.id for f in previous_files})
            or not previous_files
            and current_files
        ):
            # NOTE: Files added
            if previous_files:
                new_files = current_files - previous_files
            else:
                new_files = current_files

            descriptions.append(f"{len(new_files)} file(s) added")
            details.files_added = [str(f.entity.path) for f in new_files]
            details.modified = bool(previous_files)

        if previous_files and {f.id for f in previous_files}.difference({f.id for f in current_files}):
            # NOTE: Files removed
            removed_files = previous_files - current_files
            descriptions.append(f"{len(removed_files)} file(s) removed")
            details.files_removed = [str(f.entity.path) for f in removed_files]
            details.modified = True

        if not previous_dataset:
            # NOTE: Check metadata changes on create/import
            if dataset.title:
                details.title_changed = dataset.title

            if dataset.description:
                details.description_changed = dataset.description

            if dataset.creators:
                details.creators_added = [c.full_identity for c in dataset.creators]

            if dataset.keywords:
                details.keywords_added = [k for k in dataset.keywords]

            if dataset.images:
                details.images_changed_to = [i.content_url for i in dataset.images]
        elif not details.deleted:
            # NOTE: Check metadata changes to previous dataset
            modified = False
            if dataset.title != previous_dataset.title:
                details.title_changed = dataset.title
                modified = True
            if dataset.description != previous_dataset.description:
                details.description_changed = dataset.description
                modified = True

            current_creators = set(dataset.creators or [])
            previous_creators = set(previous_dataset.creators or [])

            if current_creators.difference(previous_creators):
                details.creators_added = [c.full_identity for c in current_creators.difference(previous_creators)]
                modified = True
            if previous_creators.difference(current_creators):
                details.creators_removed = [c.full_identity for c in previous_creators.difference(current_creators)]
                modified = True

            current_keywords = set(dataset.keywords)
            previous_keywords = set(previous_dataset.keywords)

            if current_keywords.difference(previous_keywords):
                details.keywords_added = list(current_keywords.difference(previous_keywords))
                modified = True
            if previous_keywords.difference(current_keywords):
                details.keywords_removed = list(previous_keywords.difference(current_keywords))
                modified = True

            current_images = set(dataset.images) if dataset.images else set()
            previous_images = set(previous_dataset.images) if previous_dataset.images else set()

            if current_images != previous_images:
                details.images_changed_to = [i.content_url for i in current_images]
                modified = True
            if modified:
                details.modified = True
                descriptions.append("metadata modified")

        return DatasetLogViewModel(
            id=dataset.name,
            date=dataset.date_removed
            if dataset.date_removed
            else (
                dataset.date_modified or dataset.date_created or dataset.date_published or datetime.utcfromtimestamp(0)
            ),
            description=descriptions[0] + ", ".join(descriptions[1:]),
            details=details,
            agents=[c.full_identity for c in dataset.creators],
        )


class DatasetLogViewModel(LogViewModel):
    """View model for a dataset log entry."""

    type = LogType.DATASET.value

    def __init__(
        self, id: str, date: datetime, description: str, details: DatasetChangeDetailsViewModel, agents: List[str]
    ):
        super().__init__(id, date, description, agents)
        self.details = details

    def to_dict(self) -> Dict[str, Any]:
        """Return a dict representation of this view model.

        Returns:
            Dict[str,Any]: Dictionary representation of view model.
        """
        return {
            "date": self.date.isoformat(),
            "type": self.type,
            "description": self.description,
            "agents": self.agents,
            "details": asdict(self.details),
        }


class ActivityLogViewModel(LogViewModel):
    """View model for an activity log entry."""

    type = LogType.ACTIVITY.value

    def __init__(
        self,
        id: str,
        date: datetime,
        description: str,
        details: ActivityDetailsViewModel,
        agents: List[str],
        plan: "PlanViewModel",
    ):
        super().__init__(id, date, description, agents)
        self.details = details
        self.plan = plan

    def to_dict(self) -> Dict[str, Any]:
        """Return a dict representation of this view model.

        Returns:
            Dict[str,Any]: Dictionary representation of view model.
        """
        return {
            "date": self.date.isoformat(),
            "type": self.type,
            "description": self.description,
            "agents": self.agents,
            "details": asdict(self.details),
            "plan": {"id": self.plan.id, "name": self.plan.name},
        }


LOG_COLUMNS = {
    "id": ("id", None),
    "date": ("date", None),
    "type": ("type", None),
    "description": ("description", None),
    "actors": ("agents", "actors"),
}

LOG_FORMATS = {
    "tabular": tabular,
    "json": json,
    "detailed": None,
}
